import json
import os
from typing import AsyncGenerator, List, Literal, Optional, Union, overload

from azure.ai.inference.aio import ChatCompletionsClient
from azure.ai.inference.models import (
    AssistantMessage,
    AudioContentItem,
    ChatCompletionsToolCall,
    ChatCompletionsToolDefinition,
    ChatRequestMessage,
    CompletionsFinishReason,
    ContentItem,
    FunctionCall,
    FunctionDefinition,
    ImageContentItem,
    ImageDetailLevel,
    ImageUrl,
    InputAudio,
    SystemMessage,
    TextContentItem,
    ToolMessage,
    UserMessage,
)
from azure.core.credentials import AzureKeyCredential
from azure.core.exceptions import HttpResponseError
from nonebot import logger

from ._types import (
    BasicModel,
    ModelCompletions,
    ModelConfig,
    ModelRequest,
    ModelStreamCompletions,
)
from .utils.tools import function_call_handler


class Azure(BasicModel):
    def __init__(self, model_config: ModelConfig) -> None:
        super().__init__(model_config)
        self._require("model_name")
        self.model_name = self.config.model_name
        self.max_tokens = self.config.max_tokens
        self.temperature = self.config.temperature
        self.top_p = self.config.top_p
        self.frequency_penalty = self.config.frequency_penalty
        self.presence_penalty = self.config.presence_penalty
        self.token = os.getenv("AZURE_API_KEY", self.config.api_key)
        self.endpoint = self.config.api_host if self.config.api_host else "https://models.inference.ai.azure.com"

        self._tools: List[ChatCompletionsToolDefinition] = []

    def __build_multi_messages(self, request: ModelRequest) -> UserMessage:
        """
        构建多模态类型

        此模型加载器支持的多模态类型: `audio` `image`
        """
        multi_content_items: List[ContentItem] = []

        for resource in request.resources:
            if resource.path is None:
                continue
            elif resource.type == "audio":
                multi_content_items.append(
                    AudioContentItem(
                        input_audio=InputAudio.load(audio_file=resource.path, audio_format=resource.path.split(".")[-1])
                    )
                )
            elif resource.type == "image":
                multi_content_items.append(
                    ImageContentItem(
                        image_url=ImageUrl.load(
                            image_file=resource.path,
                            image_format=resource.path.split(".")[-1],
                            detail=ImageDetailLevel.AUTO,
                        )
                    )
                )

        content = [TextContentItem(text=request.prompt)] + multi_content_items

        return UserMessage(content=content)

    def __build_tools_definition(self, tools: List[dict]) -> List[ChatCompletionsToolDefinition]:
        tool_definitions = []

        for tool in tools:
            tool_definition = ChatCompletionsToolDefinition(
                function=FunctionDefinition(
                    name=tool["function"]["name"],
                    description=tool["function"]["description"],
                    parameters=tool["function"]["parameters"],
                )
            )
            tool_definitions.append(tool_definition)

        return tool_definitions

    def _build_messages(self, request: ModelRequest) -> List[ChatRequestMessage]:
        messages: List[ChatRequestMessage] = []

        if request.system:
            messages.append(SystemMessage(request.system))

        for msg in request.history:
            user_msg = (
                UserMessage(msg.message)
                if not msg.resources
                else self.__build_multi_messages(ModelRequest(msg.message, resources=msg.resources))
            )
            messages.append(user_msg)
            messages.append(AssistantMessage(msg.respond))

        user_message = UserMessage(request.prompt) if not request.resources else self.__build_multi_messages(request)

        messages.append(user_message)

        return messages

    def _tool_messages_precheck(self, tool_calls: Optional[List[ChatCompletionsToolCall]] = None) -> bool:
        if not (tool_calls and len(tool_calls) == 1):
            return False

        tool_call = tool_calls[0]

        if isinstance(tool_call, ChatCompletionsToolCall):
            return True

        return False

    async def _ask_sync(self, messages: List[ChatRequestMessage]) -> ModelCompletions:
        client = ChatCompletionsClient(endpoint=self.endpoint, credential=AzureKeyCredential(self.token))

        completions = ModelCompletions()

        try:
            response = await client.complete(
                messages=messages,
                model=self.model_name,
                max_tokens=self.max_tokens,
                temperature=self.temperature,
                top_p=self.top_p,
                frequency_penalty=self.frequency_penalty,
                presence_penalty=self.presence_penalty,
                stream=False,
                tools=self._tools,
            )
            finish_reason = response.choices[0].finish_reason
            self._total_tokens += response.usage.total_tokens

            if finish_reason == CompletionsFinishReason.STOPPED:
                completions.text = response.choices[0].message.content

            elif finish_reason == CompletionsFinishReason.CONTENT_FILTERED:
                completions.succeed = False
                completions.text = "(模型内部错误: 被内容过滤器阻止)"

            elif finish_reason == CompletionsFinishReason.TOKEN_LIMIT_REACHED:
                completions.succeed = False
                completions.text = "(模型内部错误: 达到了最大 token 限制)"

            elif finish_reason == CompletionsFinishReason.TOOL_CALLS:
                tool_calls = response.choices[0].message.tool_calls
                messages.append(AssistantMessage(tool_calls=tool_calls))
                if (tool_calls is None) or (not self._tool_messages_precheck(tool_calls=tool_calls)):
                    completions.succeed = False
                    completions.text = "(模型内部错误: tool_calls 内容为空)"
                    return completions

                tool_call = tool_calls[0]
                function_args = json.loads(tool_call.function.arguments.replace("'", '"'))
                logger.info(f"function call 请求 {tool_call.function.name}, 参数: {function_args}")
                function_return = await function_call_handler(tool_call.function.name, function_args)
                logger.success(f"Function call 成功，返回: {function_return}")

                # Append the function call result fo the chat history
                messages.append(ToolMessage(tool_call_id=tool_call.id, content=function_return))

                return await self._ask_sync(messages)

            else:
                completions.succeed = False
                completions.text = "(模型内部错误: 达到了最大 token 限制)"

        except HttpResponseError as e:
            logger.error(f"模型响应失败: {e.status_code} ({e.reason})")
            logger.error(f"{e.message}")
            completions.succeed = False
            completions.text = f"模型响应失败: {e.status_code} ({e.reason})"

        finally:
            await client.close()
            completions.usage = self._total_tokens
            return completions

    async def _ask_stream(self, messages: List[ChatRequestMessage]) -> AsyncGenerator[ModelStreamCompletions, None]:
        client = ChatCompletionsClient(endpoint=self.endpoint, credential=AzureKeyCredential(self.token))

        try:
            response = await client.complete(
                messages=messages,
                model=self.model_name,
                max_tokens=self.max_tokens,
                temperature=self.temperature,
                top_p=self.top_p,
                frequency_penalty=self.frequency_penalty,
                presence_penalty=self.presence_penalty,
                stream=True,
                tools=self._tools,
                model_extras={"stream_options": {"include_usage": True}},  # 需要显式声明获取用量
            )

            tool_call_id: str = ""
            function_name: str = ""
            function_args: str = ""

            async for chunk in response:
                stream_completions = ModelStreamCompletions()

                if chunk.usage:  # chunk.usage 只会在最后一个包中被提供，此时choices为空
                    self._total_tokens += chunk.usage.total_tokens if chunk.usage else 0
                    stream_completions.usage = self._total_tokens

                if not chunk.choices:
                    yield stream_completions
                    continue

                finish_reason = chunk.choices[0].finish_reason

                if chunk.choices and chunk.choices[0].get("delta", {}).get("content", ""):
                    stream_completions.chunk = chunk["choices"][0]["delta"]["content"]

                elif chunk.choices[0].delta.tool_calls is not None:
                    tool_call = chunk.choices[0].delta.tool_calls[0]

                    if tool_call.function.name is not None:
                        function_name = tool_call.function.name
                    if tool_call.id is not None:
                        tool_call_id = tool_call.id
                    function_args += tool_call.function.arguments or ""
                    continue

                elif finish_reason == CompletionsFinishReason.CONTENT_FILTERED:
                    stream_completions.succeed = False
                    stream_completions.chunk = "(模型内部错误: 被内容过滤器阻止)"

                elif finish_reason == CompletionsFinishReason.TOKEN_LIMIT_REACHED:
                    stream_completions.succeed = False
                    stream_completions.chunk = "(模型内部错误: 达到了最大 token 限制)"

                elif finish_reason == CompletionsFinishReason.TOOL_CALLS:
                    messages.append(
                        AssistantMessage(
                            tool_calls=[
                                ChatCompletionsToolCall(
                                    id=tool_call_id, function=FunctionCall(name=function_name, arguments=function_args)
                                )
                            ]
                        )
                    )

                    function_arg = json.loads(function_args.replace("'", '"'))
                    logger.info(f"function call 请求 {function_name}, 参数: {function_arg}")
                    function_return = await function_call_handler(function_name, function_arg)
                    logger.success(f"Function call 成功，返回: {function_return}")

                    # Append the function call result fo the chat history
                    messages.append(ToolMessage(tool_call_id=tool_call_id, content=function_return))

                    async for content in self._ask_stream(messages):
                        yield content

                    return

                yield stream_completions

        except HttpResponseError as e:
            logger.error(f"模型响应失败: {e.status_code} ({e.reason})")
            logger.error(f"{e.message}")
            stream_completions = ModelStreamCompletions()
            stream_completions.chunk = "模型响应失败: {e.status_code} ({e.reason})"
            stream_completions.succeed = False
            yield stream_completions

        finally:
            await client.close()

    @overload
    async def ask(self, request: ModelRequest, *, stream: Literal[False] = False) -> ModelCompletions: ...

    @overload
    async def ask(
        self, request: ModelRequest, *, stream: Literal[True] = True
    ) -> AsyncGenerator[ModelStreamCompletions, None]: ...

    async def ask(
        self, request: ModelRequest, *, stream: bool = False
    ) -> Union[ModelCompletions, AsyncGenerator[ModelStreamCompletions, None]]:
        self._total_tokens = 0

        messages = self._build_messages(request)

        self._tools = self.__build_tools_definition(request.tools) if request.tools else []

        if stream:
            return self._ask_stream(messages)

        return await self._ask_sync(messages)



from nonebot import get_plugin_config
from pydantic import BaseModel


class Config(BaseModel):
    store_index: str = "https://raw.githubusercontent.com/MuikaAI/Muicebot-Plugins-Index/refs/heads/main/plugins.json"
    """插件索引文件 url"""


config = get_plugin_config(Config)



from arclet.alconna import Alconna, Subcommand
from nonebot.permission import SUPERUSER
from nonebot_plugin_alconna import Args, CommandMeta, Match, on_alconna

from muicebot.plugin import PluginMetadata

from .store import (
    get_installed_plugins_info,
    install_plugin,
    load_store_plugin,
    uninstall_plugin,
    update_plugin,
)

__meta__ = PluginMetadata(name="muicebot-plugin-store", description="Muicebot 插件商店操作", usage=".store help")

load_store_plugin()

COMMAND_PREFIXES = [".", "/"]

store_cmd = on_alconna(
    Alconna(
        COMMAND_PREFIXES,
        "store",
        Subcommand("help"),
        Subcommand("install", Args["name", str], help_text=".store install 插件名"),
        Subcommand("show"),
        Subcommand("update", Args["name", str], help_text=".store update 插件名"),
        Subcommand("uninstall", Args["name", str], help_text=".store uninstall 插件名"),
        meta=CommandMeta("Muicebot 插件商店指令"),
    ),
    priority=10,
    block=True,
    skip_for_unmatch=False,
    permission=SUPERUSER,
)


@store_cmd.assign("install")
async def install(name: Match[str]):
    if not name.available:
        await store_cmd.finish("必须传入一个插件名")
    result = await install_plugin(name.result)
    await store_cmd.finish(result)


@store_cmd.assign("show")
async def show():
    info = await get_installed_plugins_info()
    await store_cmd.finish(info)


@store_cmd.assign("update")
async def update(name: Match[str]):
    if not name.available:
        await store_cmd.finish("必须传入一个插件名")
    result = await update_plugin(name.result)
    await store_cmd.finish(result)


@store_cmd.assign("uninstall")
async def uninstall(name: Match[str]):
    if not name.available:
        await store_cmd.finish("必须传入一个插件名")
    result = await uninstall_plugin(name.result)
    await store_cmd.finish(result)


@store_cmd.assign("help")
async def store_help():
    await store_cmd.finish(
        "install <插件名> 安装插件\n"
        "show 查看已安装的商店插件信息\n"
        "update <插件名> 更新插件\n"
        "uninstall <插件名> 卸载插件"
    )



from typing import TypedDict


class PluginInfo(TypedDict):
    module: str
    """插件模块名"""
    name: str
    """插件名称"""
    description: str
    """插件描述"""
    repo: str
    """插件 repo 地址"""


class InstalledPluginInfo(TypedDict):
    module: str
    """插件模块名"""
    name: str
    """插件名称"""
    commit: str
    """commit 信息"""



from typing import List

from pydantic import BaseModel, ConfigDict, Field


class Userinfo(BaseModel):
    """用户信息配置"""

    name: str
    """用户名称"""
    id: str
    """用户 Nonebot ID"""
    info: str
    """用户信息"""


class PromptTemplatesConfig(BaseModel):
    """提示词模板配置"""

    ai_nickname: str = "沐雪"
    """AI 昵称"""
    master_nickname: str = "沐沐(Muika)"
    """AI 开发者昵称"""

    userinfos: List[Userinfo] = Field(default=[])
    """用户信息列表"""

    model_config = ConfigDict(extra="allow")
    """允许其他模板参数传入"""


class PromptTemplatesData(BaseModel):
    """提示词模板数据"""

    ai_nickname: str = "沐雪"
    """AI 昵称"""
    master_nickname: str = "沐沐(Muika)"
    """AI 开发者昵称"""

    private: bool = False
    """当前对话是否为私聊"""
    user_name: str = ""
    """目标用户名"""
    user_info: str = ""
    """目标用户信息"""

    model_config = ConfigDict(extra="allow")
    """允许其他模板参数传入"""

    @classmethod
    def from_config(cls, templates_config: PromptTemplatesConfig, userid: str, is_private: bool = False):
        base = templates_config.model_dump()
        data = cls(**base)

        user = next((u for u in templates_config.userinfos if u.id == userid), None)
        if user:
            data.user_name = user.name
            data.user_info = user.info

        data.private = is_private

        return data



from __future__ import annotations

import json
import shutil
from pathlib import Path
from typing import TYPE_CHECKING

import aiosqlite
from nonebot import logger

if TYPE_CHECKING:
    from ..database import Database


class MigrationManager:
    """数据库迁移管理器"""

    def __init__(self, db: Database) -> None:
        self.db = db
        self.path: Path = db.DB_PATH

        self.migrations = {0: self._migrate_v0_to_v1, 1: self._migrate_v1_to_v2}

        self.latest_version = max(self.migrations.keys()) + 1

    async def migrate_if_needed(self):
        """
        检查数据库更新并迁移
        """
        await self.__init_version_table()
        current_version = await self.__get_version()

        while current_version in self.migrations:
            logger.info(f"检测到数据库更新，当前版本 v{current_version}...")
            backup_path = self.path.with_suffix(f".backup.v{current_version}.db")
            shutil.copyfile(self.path, backup_path)

            try:
                async with self.db.connect() as conn:
                    await conn.execute("BEGIN")
                    await self.migrations[current_version](conn)
                    await conn.commit()
                current_version += 1
                await self.__set_version(current_version)
                logger.success(f"数据库已成功迁移到 v{current_version} ⭐")
            except Exception as e:
                logger.error(f"迁移至 v{current_version + 1} 失败，错误：{e}")
                shutil.copyfile(backup_path, self.path)
                logger.info("已回退到迁移前的状态")
                break

    async def __init_version_table(self):
        await self.db.execute(
            """
            CREATE TABLE IF NOT EXISTS schema_version (
                version INTEGER NOT NULL
            );
        """
        )
        result = await self.db.execute("SELECT COUNT(*) FROM schema_version", fetchone=True)
        if not result or result[0] == 0:  # 只有 v0.2.x 的数据库没有版本号
            await self.db.execute("INSERT INTO schema_version (version) VALUES (0)")
            logger.info("数据库无版本记录，可能是v0版本的数据库，设为v0")

    async def __get_version(self) -> int:
        """
        获取数据库版本号，默认值为0
        """
        result = await self.db.execute("SELECT version FROM schema_version", fetchone=True)
        return result[0] if result else 0

    async def __set_version(self, version: int):
        await self.db.execute("UPDATE schema_version SET version = ?", (version,))

    async def _migrate_v0_to_v1(self, conn: aiosqlite.Connection):
        logger.info("v1 更新内容: 添加 TOTALTOKENS 与 GROUPID 字段")
        await conn.execute("ALTER TABLE MSG ADD COLUMN TOTALTOKENS INTEGER DEFAULT -1;")
        await conn.execute("ALTER TABLE MSG ADD COLUMN GROUPID TEXT DEFAULT '-1';")

    async def _migrate_v1_to_v2(self, conn: aiosqlite.Connection):
        logger.info("v2 更新内容: total_token 变更为 usage, images 列表优化为 resources")

        # 创建临时表
        await conn.execute(
            """
        CREATE TABLE MSG_NEW(
            ID INTEGER PRIMARY KEY AUTOINCREMENT,
            TIME TEXT NOT NULL,
            USERID TEXT NOT NULL,
            GROUPID TEXT NOT NULL DEFAULT (-1),
            MESSAGE TEXT NOT NULL,
            RESPOND TEXT NOT NULL,
            HISTORY INTEGER NOT NULL DEFAULT (1),
            RESOURCES TEXT NOT NULL DEFAULT "[]",
            USAGE INTEGER NOT NULL DEFAULT (-1)
        );
        """
        )

        cursor = await conn.execute(
            "SELECT ID, TIME, USERID, GROUPID, MESSAGE, RESPOND, HISTORY, IMAGES, TOTALTOKENS FROM MSG"
        )
        rows = await cursor.fetchall()

        for row in rows:
            id_, time_, userid, groupid, message, respond, history, images_json, totaltokens = row

            # 转换 images -> resources
            try:
                images = json.loads(images_json) if images_json else []
            except Exception:
                images = []

            resources = []
            for url in images:
                resources.append({"type": "image", "path": url})
            resources_json = json.dumps(resources, ensure_ascii=False)

            # 插入到新表
            await conn.execute(
                """
            INSERT INTO MSG_NEW (ID, TIME, USERID, GROUPID, MESSAGE, RESPOND, HISTORY, RESOURCES, USAGE)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
            """,
                (id_, time_, userid, groupid, message, respond, history, resources_json, totaltokens),
            )

        await conn.execute("DROP TABLE MSG")
        await conn.execute("ALTER TABLE MSG_NEW RENAME TO MSG")



import base64
import mimetypes
import os
import ssl
import sys
import time
from importlib.metadata import PackageNotFoundError, version
from typing import Optional

import fleep
import httpx
import nonebot_plugin_localstore as store
from nonebot import get_bot, logger
from nonebot.adapters import Event, MessageSegment
from nonebot.log import default_filter, logger_id
from nonebot_plugin_userinfo import get_user_info

from ..config import plugin_config
from ..models import Resource
from ..plugin.context import get_event
from .adapters import ADAPTER_CLASSES

IMG_DIR = store.get_plugin_data_dir() / ".cache" / "images"
IMG_DIR.mkdir(parents=True, exist_ok=True)

User_Agent = (
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"
    "AppleWebKit/537.36 (KHTML, like Gecko)"
    "Chrome/134.0.0.0 Safari/537.36 Edg/134.0.0.0"
)


async def download_file(file_url: str, file_name: Optional[str] = None, proxy: Optional[str] = None) -> str:
    """
    保存文件至本地目录

    :param file_url: 图片在线地址
    :param file_name: 要保存的文件名
    :param proxy: 代理地址

    :return: 保存后的本地目录
    """
    ssl_context = ssl.create_default_context()
    ssl_context.set_ciphers("DEFAULT")
    file_name = file_name if file_name else str(time.time_ns()) + ".jpg"

    async with httpx.AsyncClient(proxy=proxy, verify=ssl_context) as client:
        r = await client.get(file_url, headers={"User-Agent": User_Agent})
        local_path = (IMG_DIR / file_name).resolve()
        with open(local_path, "wb") as file:
            file.write(r.content)
        return str(local_path)


async def save_image_as_base64(image_url: str, proxy: Optional[str] = None) -> str:
    """
    从在线 url 获取图像 Base64

    :image_url: 图片在线地址
    :return: 本地地址
    """
    ssl_context = ssl.create_default_context()
    ssl_context.set_ciphers("DEFAULT")

    async with httpx.AsyncClient(proxy=proxy, verify=ssl_context) as client:
        r = await client.get(image_url, headers={"User-Agent": User_Agent})
        image_base64 = base64.b64encode(r.content)
    return image_base64.decode("utf-8")


async def get_file_via_adapter(message: MessageSegment, event: Event) -> Optional[str]:
    """
    通过适配器自有方式获取文件地址并保存到本地

    :return: 本地地址
    """
    bot = get_bot()

    Onebotv12Bot = ADAPTER_CLASSES["onebot_v12"]
    UnsupportedParam = ADAPTER_CLASSES["UnsupportedParam"]
    Onebotv11Bot = ADAPTER_CLASSES["onebot_v11"]
    TelegramEvent = ADAPTER_CLASSES["telegram_event"]
    TelegramFile = ADAPTER_CLASSES["telegram_file"]

    if Onebotv12Bot and UnsupportedParam and isinstance(bot, Onebotv12Bot):
        # if message.type != "image":
        #     return None

        try:
            file_path = await bot.get_file(type="url", file_id=message.data["file_id"])
        except UnsupportedParam as e:
            logger.error(f"Onebot 实现不支持获取文件 URL，文件获取操作失败：{e}")
            return None

        return str(file_path)

    elif Onebotv11Bot and isinstance(bot, Onebotv11Bot):
        if "url" in message.data and "file" in message.data:
            return await download_file(message.data["url"], message.data["file"])

    elif TelegramEvent and TelegramFile and isinstance(event, TelegramEvent):
        if not isinstance(message, TelegramFile):
            return None

        file_id = message.data["file"]
        file = await bot.get_file(file_id=file_id)
        if not file.file_path:
            return None

        url = f"https://api.telegram.org/file/bot{bot.bot_config.token}/{file.file_path}"  # type: ignore
        # filename = file.file_path.split("/")[1]
        return await download_file(url, proxy=plugin_config.telegram_proxy)

    return None


def init_logger():
    console_handler_level = plugin_config.log_level

    log_dir = "logs"
    if not os.path.exists(log_dir):
        os.mkdir(log_dir)

    log_file_path = f"{log_dir}/{time.strftime('%Y-%m-%d')}.log"

    # 移除 NoneBot 默认的日志处理器
    logger.remove(logger_id)
    # 添加新的日志处理器
    logger.add(
        sys.stdout,
        level=console_handler_level,
        diagnose=True,
        format="<lvl>[{level}] {function}: {message}</lvl>",
        filter=default_filter,
        colorize=True,
    )

    logger.add(
        log_file_path,
        level="DEBUG",
        format="[{time:YYYY-MM-DD HH:mm:ss}] [{level}] {function}: {message}",
        encoding="utf-8",
        rotation="1 day",
        retention="7 days",
    )


def get_version() -> str:
    """
    获取当前版本号

    优先尝试从已安装包中获取版本号, 否则从 `pyproject.toml` 读取
    """
    package_name = "muicebot"

    try:
        return version(package_name)
    except PackageNotFoundError:
        pass

    toml_path = os.path.join(os.path.dirname(__file__), "../pyproject.toml")

    if not os.path.isfile(toml_path):
        return "Unknown"

    try:
        if sys.version_info >= (3, 11):
            import tomllib

            with open(toml_path, "rb") as f:
                pyproject_data = tomllib.load(f)

        else:
            import toml

            with open(toml_path, "r", encoding="utf-8") as f:
                pyproject_data = toml.load(f)

        # 返回版本号
        return pyproject_data["tool"]["pdm"]["version"]

    except (FileNotFoundError, KeyError, ModuleNotFoundError):
        return "Unknown"


async def get_username(user_id: Optional[str] = None) -> str:
    """
    获取当前对话的用户名，如果失败就返回用户id
    """
    bot = get_bot()
    event = get_event()
    user_id = user_id if user_id else event.get_user_id()
    user_info = await get_user_info(bot, event, user_id)
    return user_info.user_name if user_info else user_id


def guess_mimetype(resource: Resource) -> Optional[str]:
    """
    尝试获取 minetype 类型
    """
    # raw 不落库，因此无法从 raw 判断
    if resource.path and os.path.exists(resource.path):
        try:
            with open(resource.path, "rb") as file:
                header = file.read(128)
                info = fleep.get(header)
                if info.mime:
                    return info.mime[0]  # type:ignore
                else:
                    return mimetypes.guess_type(resource.path)[0]
        except Exception:
            return mimetypes.guess_type(resource.path)[0]
    elif resource.url:
        return mimetypes.guess_type(resource.url)[0]
    return None



import base64
from typing import Optional


def get_file_base64(local_path: Optional[str] = None, file_bytes: Optional[bytes] = None) -> str:
    """
    获取本地图像 Base64 的方法
    """
    if local_path:
        with open(local_path, "rb") as f:
            image_data = base64.b64encode(f.read()).decode("utf-8")
            return image_data
    if file_bytes:
        image_base64 = base64.b64encode(file_bytes)
        return image_base64.decode("utf-8")
    raise ValueError("You must pass in a valid parameter!")



"""
存储并获取 Nonebot 依赖注入中的上下文
"""

from contextvars import ContextVar
from typing import Tuple

from nonebot.adapters import Bot, Event
from nonebot.matcher import Matcher
from nonebot.typing import T_State

# 定义上下文变量
bot_context: ContextVar[Bot] = ContextVar("bot")
event_context: ContextVar[Event] = ContextVar("event")
state_context: ContextVar[T_State] = ContextVar("state")
mather_context: ContextVar[Matcher] = ContextVar("matcher")


# 获取当前上下文的各种信息
def get_bot() -> Bot:
    return bot_context.get()


def get_event() -> Event:
    return event_context.get()


def get_state() -> T_State:
    return state_context.get()


def get_mather() -> Matcher:
    return mather_context.get()


def set_ctx(bot: Bot, event: Event, state: T_State, matcher: Matcher):
    """
    注册 Nonebot 中的上下文信息
    """
    bot_context.set(bot)
    event_context.set(event)
    state_context.set(state)
    mather_context.set(matcher)


def get_ctx() -> Tuple[Bot, Event, T_State, Matcher]:
    """
    获取当前上下文
    """
    return (get_bot(), get_event(), get_state(), get_mather())



"""
机器人框架中处理函数调用的模块。
该模块提供了一个系统，用于注册可被AI系统调用的函数，
具有自动依赖注入和参数验证功能。
它包括：
- Caller类：管理函数注册、依赖注入和执行
- 用于函数调用的注册装饰器
- 用于获取已注册函数调用的实用函数
"""

import inspect
from typing import Any, Optional, get_type_hints

from nonebot import logger
from nonebot.adapters import Bot, Event
from nonebot.matcher import Matcher
from nonebot.rule import Rule
from nonebot.typing import T_State

from ..context import get_bot, get_event, get_mather
from ..utils import is_coroutine_callable
from ._types import ASYNC_FUNCTION_CALL_FUNC, F
from .parameter import Parameter
from .utils import async_wrap

_caller_data: dict[str, "Caller"] = {}
"""函数注册表，存储所有注册的函数"""


class Caller:
    def __init__(self, description: str, rule: Optional[Rule] = None):
        self._name: str = ""
        """函数名称"""
        self._description: str = description
        """函数描述"""
        self._rule: Optional[Rule] = rule
        """启用规则"""
        self._parameters: dict[str, Parameter] = {}
        """函数参数字典"""
        self.function: ASYNC_FUNCTION_CALL_FUNC
        """函数对象"""
        self.default: dict[str, Any] = {}
        """默认值"""

        self.module_name: str = ""
        """函数所在模块名称"""

    def __call__(self, func: F) -> F:
        """
        修饰器：注册一个 Function_call 函数
        """
        # 确保为异步函数
        if is_coroutine_callable(func):
            self.function = func  # type: ignore
        else:
            self.function = async_wrap(func)  # type:ignore

        self._name = func.__name__

        # 获取模块名
        if module := inspect.getmodule(func):
            module_name = module.__name__.split(".")[-1]
        else:
            module_name = ""
        self.module_name = module_name

        _caller_data[self._name] = self
        logger.success(f"Function Call 函数 {self.module_name}.{self._name} 已成功加载")
        return func

    async def _inject_dependencies(self, kwargs: dict) -> dict:
        """
        自动解析参数并进行依赖注入
        """
        sig = inspect.signature(self.function)
        hints = get_type_hints(self.function)

        inject_args = kwargs.copy()

        for name, param in sig.parameters.items():
            param_type = hints.get(name, None)

            if param_type and isinstance(param_type, type) and issubclass(param_type, Bot):
                inject_args[name] = get_bot()

            elif param_type and isinstance(param_type, type) and issubclass(param_type, Event):
                inject_args[name] = get_event()

            elif param_type and isinstance(param_type, type) and issubclass(param_type, Matcher):
                inject_args[name] = get_mather()

            # elif param_type and issubclass(param_type, T_State):
            #     inject_args[name] = get_state()

            # 填充默认值
            elif param.default != inspect.Parameter.empty:
                inject_args[name] = kwargs.get(name, param.default)

            # 如果参数未提供，则检查是否有默认值
            elif name not in inject_args:
                raise ValueError(f"缺少必要参数: {name}")

        return inject_args

    def params(self, **kwargs: Parameter) -> "Caller":
        self._parameters.update(kwargs)
        return self

    async def run(self, **kwargs) -> Any:
        """
        执行 function call
        """
        if self.function is None:
            raise ValueError("未注册函数对象")

        inject_args = await self._inject_dependencies(kwargs)

        return await self.function(**inject_args)

    def data(self) -> dict[str, Any]:
        """
        生成函数描述信息

        :return: 可用于 Function_call 的字典
        """
        if not self._parameters:
            properties = {
                "dummy_param": {"type": "string", "description": "为了兼容性设置的一个虚拟参数，因此不需要填写任何值"}
            }
            required = []
        else:
            properties = {key: value.data() for key, value in self._parameters.items()}
            required = [key for key, value in self._parameters.items() if value.default is None]

        return {
            "type": "function",
            "function": {
                "name": self._name,
                "description": self._description,
                "parameters": {
                    "type": "object",
                    "properties": properties,
                    "required": required,
                },
            },
        }


def on_function_call(description: str, rule: Optional[Rule] = None) -> Caller:
    """
    返回一个Caller类，可用于装饰一个函数，使其注册为一个可被AI调用的function call函数

    :param description: 函数描述，若为None则从函数的docstring中获取
    :param rule: 启用规则。不满足规则则不启用此 function call

    :return: Caller对象
    """
    caller = Caller(description=description, rule=rule)
    return caller


def get_function_calls() -> dict[str, Caller]:
    """获取所有已注册的function call函数

    Returns:
        dict[str, Caller]: 所有已注册的function call类
    """
    return _caller_data


async def get_function_list() -> list[dict[str, dict]]:
    """
    获取所有已注册的function call函数，并转换为工具格式

    :return: 所有已注册的function call函数列表
    """
    tools: list[dict[str, dict]] = []
    bot: Bot = get_bot()
    event: Event = get_event()
    state: T_State = {}

    for name, caller in _caller_data.items():
        if caller._rule is None or await caller._rule(bot, event, state):
            tools.append(caller.data())

    return tools



import asyncio
from typing import Any, Optional

from nonebot import logger

from .config import server_config
from .server import Server, Tool

_servers: list[Server] = list()


async def initialize_servers() -> None:
    """
    初始化全部 MCP 实例
    """
    _servers.extend([Server(name, srv_config) for name, srv_config in server_config.items()])
    for server in _servers:
        logger.info(f"初始化 MCP Server: {server.name}")
        try:
            await server.initialize()
        except Exception as e:
            logger.error(f"初始化 MCP Server 实例时出现问题: {e}")
            await cleanup_servers()
            raise


async def handle_mcp_tool(tool: str, arguments: Optional[dict[str, Any]] = None) -> Optional[str]:
    """
    处理 MCP Tool 调用
    """
    logger.info(f"执行 MCP 工具: {tool} (参数: {arguments})")

    for server in _servers:
        server_tools = await server.list_tools()
        if not any(server_tool.name == tool for server_tool in server_tools):
            continue

        try:
            result = await server.execute_tool(tool, arguments)

            if isinstance(result, dict) and "progress" in result:
                progress = result["progress"]
                total = result["total"]
                percentage = (progress / total) * 100
                logger.info(f"工具执行进度: {progress}/{total} ({percentage:.1f}%)")

            return f"Tool execution result: {result}"
        except Exception as e:
            error_msg = f"Error executing tool: {str(e)}"
            logger.error(error_msg)
            return error_msg

    return None  # Not found.


async def cleanup_servers() -> None:
    """
    清理 MCP 实例
    """
    cleanup_tasks = [asyncio.create_task(server.cleanup()) for server in _servers]
    if cleanup_tasks:
        try:
            await asyncio.gather(*cleanup_tasks, return_exceptions=True)
        except Exception as e:
            logger.warning(f"清理 MCP 实例时出现错误: {e}")


async def transform_json(tool: Tool) -> dict[str, Any]:
    """
    将 MCP Tool 转换为 OpenAI 所需的 parameters 格式，并删除多余字段
    """
    func_desc = {"name": tool.name, "description": tool.description, "parameters": {}, "required": []}

    if tool.input_schema:
        parameters = {
            "type": tool.input_schema.get("type", "object"),
            "properties": tool.input_schema.get("properties", {}),
            "required": tool.input_schema.get("required", []),
        }
        func_desc["parameters"] = parameters

    output = {"type": "function", "function": func_desc}

    return output


async def get_mcp_list() -> list[dict[str, dict]]:
    """
    获得适用于 OpenAI Tool Call 输入格式的 MCP 工具列表
    """
    all_tools: list[dict[str, dict]] = []

    for server in _servers:
        tools = await server.list_tools()
        all_tools.extend([await transform_json(tool) for tool in tools])

    return all_tools



from .context import get_bot, get_ctx, get_event, get_mather, get_state, set_ctx
from .loader import (
    get_plugin_by_module_name,
    get_plugin_data_dir,
    get_plugins,
    load_plugin,
    load_plugins,
)
from .models import Plugin, PluginMetadata

__all__ = [
    "get_bot",
    "get_state",
    "get_ctx",
    "get_event",
    "get_mather",
    "load_plugin",
    "load_plugins",
    "get_plugins",
    "get_plugin_by_module_name",
    "PluginMetadata",
    "Plugin",
    "set_ctx",
    "get_plugin_data_dir",
]



import json
from pathlib import Path

from .models import InstalledPluginInfo

JSON_PATH = Path("plugins/installed_plugins.json")


def load_json_record() -> dict[str, InstalledPluginInfo]:
    """
    获取本地 json 记录
    `plugin/installed_plugins.json`
    """
    if not JSON_PATH.exists():
        return {}

    try:
        with open(JSON_PATH, "r", encoding="utf-8") as f:
            return json.load(f)
    except Exception:
        return {}


def _save_json_record(data: dict) -> None:
    """
    在本地保存 json 记录
    """
    with open(JSON_PATH, "w", encoding="utf-8") as f:
        json.dump(data, f, indent=2, ensure_ascii=False)


def register_plugin(plugin: str, commit: str, name: str, module: str) -> None:
    """
    在本地注册一个插件记录

    :param plugin: 插件唯一索引名
    :param commit: 插件 git commit hash值
    :param name: 插件名
    :param module: 相对于 `store` 文件夹的可导入模块名
    """
    plugins = load_json_record()
    plugins[plugin] = {"commit": commit, "module": module, "name": name}
    _save_json_record(plugins)


def unregister_plugin(plugin: str) -> None:
    """
    取消记录一个插件记录（通常在卸载插件时使用）
    """
    plugins = load_json_record()
    if plugin in plugins:
        del plugins[plugin]
        _save_json_record(plugins)



import re
import time
from datetime import timedelta
from pathlib import Path
from typing import AsyncGenerator, Literal

import nonebot_plugin_localstore as store
from arclet.alconna import Alconna, AllParam, Args
from nonebot import (
    get_adapters,
    get_driver,
    logger,
)
from nonebot.adapters import Bot, Event
from nonebot.adapters import Message as BotMessage
from nonebot.matcher import Matcher
from nonebot.permission import SUPERUSER
from nonebot.rule import to_me
from nonebot.typing import T_State
from nonebot_plugin_alconna import (
    AlconnaMatch,
    CommandMeta,
    Match,
    MsgTarget,
    UniMessage,
    on_alconna,
    uniseg,
)
from nonebot_plugin_alconna.builtins.extensions import ReplyRecordExtension
from nonebot_plugin_alconna.uniseg import UniMsg
from nonebot_plugin_session import SessionIdType, extract_session

from .config import plugin_config
from .llm import ModelCompletions, ModelStreamCompletions
from .models import Message, Resource
from .muice import Muice
from .plugin import get_plugins, load_plugins, set_ctx
from .plugin.mcp import initialize_servers
from .scheduler import setup_scheduler
from .utils.SessionManager import SessionManager
from .utils.utils import download_file, get_file_via_adapter, get_version

COMMAND_PREFIXES = [".", "/"]
PLUGINS_PATH = Path("./plugins")
MCP_CONFIG_PATH = Path("./configs/mcp.json")
START_TIME = time.time()

muice = Muice()
scheduler = None
connect_time = 0.0
session_manager = SessionManager()

muice_nicknames = plugin_config.muice_nicknames
regex_patterns = [f"^{re.escape(nick)}\\s*" for nick in muice_nicknames]
combined_regex = "|".join(regex_patterns)

driver = get_driver()
adapters = get_adapters()


@driver.on_startup
async def load_bot():
    logger.info(f"MuiceBot 版本: {get_version()}")
    logger.info(f"MuiceBot 数据目录: {store.get_plugin_data_dir().resolve()}")
    logger.info("加载 MuiceBot 框架...")

    logger.info(f"加载模型适配器: {muice.model_loader} ...")
    if not muice.load_model():
        logger.error("模型加载失败，请检查配置项是否正确")
        exit(1)
    logger.success(f"模型适配器加载成功: {muice.model_loader} ⭐")

    if PLUGINS_PATH.exists():
        logger.info("加载外部插件...")
        load_plugins("./plugins")

    if plugin_config.enable_builtin_plugins:
        logger.info("加载 MuiceBot 内嵌插件...")
        builtin_plugins_path = Path(__file__).parent / "builtin_plugins"
        muicebot_plugins_path = Path(__file__).resolve().parent.parent
        load_plugins(builtin_plugins_path, base_path=muicebot_plugins_path)

    if MCP_CONFIG_PATH.exists():
        logger.info("加载 MCP Server 配置")
        await initialize_servers()

    logger.success("插件加载完成⭐")

    logger.success("MuiceBot 已准备就绪✨")


@driver.on_bot_connect
async def bot_connected():
    logger.success("Bot 已连接，消息处理进程开始运行✨")
    global connect_time
    if not connect_time:
        connect_time = time.time()


command_about = on_alconna(
    Alconna(COMMAND_PREFIXES, "about", meta=CommandMeta("输出关于信息")),
    priority=90,
    block=True,
)

command_help = on_alconna(
    Alconna(COMMAND_PREFIXES, "help", meta=CommandMeta("输出帮助信息")),
    priority=90,
    block=True,
)

command_status = on_alconna(
    Alconna(COMMAND_PREFIXES, "status", meta=CommandMeta("显示当前状态")),
    priority=90,
    block=True,
)

command_reset = on_alconna(
    Alconna(COMMAND_PREFIXES, "reset", meta=CommandMeta("清空对话记录")),
    priority=10,
    block=True,
)

command_refresh = on_alconna(
    Alconna(COMMAND_PREFIXES, "refresh", meta=CommandMeta("刷新模型输出")),
    priority=10,
    block=True,
)

command_undo = on_alconna(
    Alconna(COMMAND_PREFIXES, "undo", meta=CommandMeta("撤回上一个对话")),
    priority=10,
    block=True,
)

command_load = on_alconna(
    Alconna(
        COMMAND_PREFIXES,
        "load",
        Args["config_name", str, ""],
        meta=CommandMeta("加载模型", usage="load <config_name>", example="load model.deepseek"),
    ),
    priority=10,
    block=True,
    permission=SUPERUSER,
)

command_schedule = on_alconna(
    Alconna(COMMAND_PREFIXES, "schedule", meta=CommandMeta("加载定时任务")),
    priority=10,
    block=True,
    permission=SUPERUSER,
)

command_start = on_alconna(
    Alconna(COMMAND_PREFIXES, "start", meta=CommandMeta("Telegram 的启动指令")),
    priority=10,
    block=True,
)

command_whoami = on_alconna(
    Alconna(COMMAND_PREFIXES, "whoami", meta=CommandMeta("输出当前用户信息")),
    priority=90,
    block=True,
)

nickname_event = on_alconna(
    Alconna(re.compile(combined_regex), Args["text?", AllParam], separators=""),
    priority=99,
    block=True,
    extensions=[ReplyRecordExtension()],
)

at_event = on_alconna(
    Alconna(re.compile(".+"), Args["text?", AllParam], separators=""),
    priority=100,
    rule=to_me(),
    block=True,
    extensions=[ReplyRecordExtension()],
)


@driver.on_bot_connect
@command_schedule.handle()
async def on_bot_connect():
    global scheduler
    if not scheduler:
        scheduler = setup_scheduler(muice)


@driver.on_bot_disconnect
async def on_bot_disconnect():
    global scheduler
    if scheduler:
        scheduler.remove_all_jobs()
        scheduler = None


@command_help.handle()
async def handle_command_help():
    await command_help.finish(
        "基本命令：\n"
        "about 获取关于信息\n"
        "help 输出此帮助信息\n"
        "status 显示当前状态\n"
        "refresh 刷新模型输出\n"
        "reset 清空对话记录\n"
        "undo 撤回上一个对话\n"
        "whoami 输出当前用户信息\n"
        "load <config_name> 加载模型\n"
        "（支持的命令前缀：“.”、“/”）"
    )


@command_about.handle()
async def handle_command_about():
    model_loader = muice.model_loader
    # plugins_list = ", ".join(get_available_plugin_names())
    mplugins_list = ", ".join(get_plugins())

    model_name = muice.model_config.model_name if muice.model_config.model_name else "Unknown"
    is_multimodal = "是" if muice.multimodal else "否"

    if scheduler and scheduler.running:
        job_ids = [job.id for job in scheduler.get_jobs()]
        if job_ids:
            current_scheduler = ", ".join(job_ids)
        else:
            current_scheduler = "无"
    else:
        current_scheduler = "无(调度器未启动)"

    await command_about.finish(
        f"框架版本: {get_version()}\n"
        f"已加载的 Muicebot 插件: {mplugins_list}\n"
        f"\n"
        f"模型: {model_name}({model_loader})\n"
        f"多模态: {is_multimodal}\n"
        f"\n"
        f"定时任务: {current_scheduler}"
    )


@command_status.handle()
async def handle_command_status():
    now = time.time()
    uptime = timedelta(seconds=int(now - START_TIME))
    bot_uptime = timedelta(seconds=int(now - connect_time))

    model_status = "运行中" if muice.model and muice.model.is_running else "未启动"
    today_usage, total_usage = await muice.database.get_model_usage()

    scheduler_status = "运行中" if scheduler and scheduler.running else "未启动"

    await command_status.finish(
        f"框架已运行: {str(uptime)}\n"
        f"bot已稳定连接: {str(bot_uptime)}\n"
        f"\n"
        f"模型加载器状态: {model_status}\n"
        f"今日模型用量: {today_usage} tokens (总 {total_usage} tokens)\n "
        f"\n"
        f"定时任务调度器状态: {scheduler_status}\n"
    )


@command_reset.handle()
async def handle_command_reset(event: Event):
    userid = event.get_user_id()
    response = await muice.reset(userid)
    await command_reset.finish(response)


@command_refresh.handle()
async def handle_command_refresh(bot: Bot, event: Event, state: T_State, matcher: Matcher):
    userid = event.get_user_id()

    set_ctx(bot, event, state, matcher)

    response = await muice.refresh(userid)

    await _send_message(response)


@command_undo.handle()
async def handle_command_undo(event: Event):
    userid = event.get_user_id()
    response = await muice.undo(userid)
    await command_undo.finish(response)


@command_load.handle()
async def handle_command_load(config: Match[str] = AlconnaMatch("config_name")):
    config_name = config.result
    result = muice.change_model_config(config_name)
    await UniMessage(result).finish()


@command_whoami.handle()
async def handle_command_whoami(bot: Bot, event: Event):
    user_id = event.get_user_id()
    session = extract_session(bot, event)
    group_id = session.get_id(SessionIdType.GROUP)
    session_id = event.get_session_id()
    await UniMessage(f"用户 ID: {user_id}\n群组 ID: {group_id}\n当前会话信息: {session_id}").finish()


@command_start.handle()
async def handle_command_start():
    pass


async def _extract_multi_resource(
    message: UniMessage, type: Literal["audio", "image", "video", "file"], event: Event
) -> list[Resource]:
    """
    提取单个多模态文件
    """
    resources = []

    for resource in message:
        assert isinstance(resource, uniseg.segment.Media)  # 正常情况下应该都是 Media 的子类

        try:
            if resource.path is not None:
                path = str(resource.path)
            elif resource.url is not None:
                path = await download_file(resource.url, file_name=resource.name)
            elif resource.origin is not None:
                logger.warning("无法通过通用方式获取文件URL，回退至适配器自有方式...")
                path = await get_file_via_adapter(resource.origin, event)  # type:ignore
            else:
                continue

            if path:
                resources.append(Resource(type, path=path))
        except Exception as e:
            logger.error(f"处理文件失败: {e}")

    return resources


async def _extract_multi_resources(message: UniMsg, event: Event) -> list[Resource]:
    """
    提取多个多模态文件
    """
    if not muice.model_config.multimodal:
        return []

    resources = []

    message_audio = message.get(uniseg.Audio) + message.get(uniseg.Voice)
    message_images = message.get(uniseg.Image)
    message_file = message.get(uniseg.File)
    message_video = message.get(uniseg.Video)

    resources.extend(await _extract_multi_resource(message_audio, "audio", event))
    resources.extend(await _extract_multi_resource(message_file, "file", event))
    resources.extend(await _extract_multi_resource(message_images, "image", event))
    resources.extend(await _extract_multi_resource(message_video, "video", event))

    return resources


async def _send_multi_messages(resource: Resource):
    """
    发送多模态文件

    TODO: 我们有可能对发送对象添加文件名吗？
    """
    if resource.type == "audio":
        await UniMessage(uniseg.Voice(raw=resource.raw, path=resource.path)).send()
    elif resource.type == "image":
        await UniMessage(uniseg.Image(raw=resource.raw, path=resource.path)).send()
    elif resource.type == "video":
        await UniMessage(uniseg.Video(raw=resource.raw, path=resource.path)).send()
    else:
        await UniMessage(uniseg.File(raw=resource.raw, path=resource.path)).send()


async def _send_message(completions: ModelCompletions | AsyncGenerator[ModelStreamCompletions, None]):
    # non-stream
    if isinstance(completions, ModelCompletions):
        paragraphs = completions.text.split("\n\n")

        for index, paragraph in enumerate(paragraphs):
            if not paragraph.strip():
                continue  # 跳过空白文段
            if index == len(paragraphs) - 1:
                await UniMessage(paragraph).finish()
            await UniMessage(paragraph).send()

        if completions.resources:
            for resource in completions.resources:
                await _send_multi_messages(resource)

        return

    # stream
    current_paragraph = ""

    async for chunk in completions:
        logger.debug(chunk)
        current_paragraph += chunk.chunk
        paragraphs = current_paragraph.split("\n\n")

        while len(paragraphs) > 1:
            current_paragraph = paragraphs[0].strip()
            if current_paragraph:
                await UniMessage(current_paragraph).send()
            paragraphs = paragraphs[1:]

        current_paragraph = paragraphs[-1].strip()

        if chunk.resources:
            for resource in chunk.resources:
                await _send_multi_messages(resource)

    if current_paragraph:
        await UniMessage(current_paragraph).finish()


@at_event.handle()
@nickname_event.handle()
async def handle_supported_adapters(
    bot_message: UniMsg,
    event: Event,
    bot: Bot,
    state: T_State,
    matcher: Matcher,
    target: MsgTarget,
    ext: ReplyRecordExtension,
):
    if any((bot_message.startswith("."), bot_message.startswith("/"))):
        await UniMessage("未知的指令或权限不足").finish()

    # 先拿到引用消息并合并到 message (如果有)
    if message_reply := ext.get_reply(bot_message.get_message_id()):
        reply_message = message_reply.msg
        if isinstance(reply_message, BotMessage):
            bot_message += UniMessage("\n被引用的消息: ") + await UniMessage.generate(message=reply_message)
        else:
            bot_message += UniMessage(f"\n被引用的消息: {reply_message}")

    # 然后等待新消息插入
    if not (merged_message := await session_manager.put_and_wait(event, bot_message)):
        matcher.skip()
        return  # 防止类型检查器错误推断 merged_message 类型)

    message_text = merged_message.extract_plain_text()
    message_resource = await _extract_multi_resources(merged_message, event)

    userid = event.get_user_id()
    if not target.private:
        session = extract_session(bot, event)
        group_id = session.get_id(SessionIdType.GROUP)
    else:
        group_id = "-1"

    set_ctx(bot, event, state, matcher)  # 注册上下文信息以供插件、传统图片获取器使用

    logger.info(f"收到消息文本: {message_text} 多模态消息: {message_resource}")

    if not any((message_text, message_resource)):
        return

    message = Message(message=message_text, userid=userid, groupid=group_id, resources=message_resource)

    # Stream
    if muice.model_config.stream:
        stream_completions = muice.ask_stream(message)
        await _send_message(stream_completions)
        return

    # non-stream
    completions = await muice.ask(message)

    logger.info(f"生成最终回复: {completions}")

    await _send_message(completions)



from typing import Any

from muicebot.plugin.func_call import get_function_calls
from muicebot.plugin.mcp import handle_mcp_tool


async def function_call_handler(func: str, arguments: dict[str, str] | None = None) -> Any:
    """
    模型 Function Call 请求处理
    """
    arguments = arguments if arguments and arguments != {"dummy_param": ""} else {}

    if func_caller := get_function_calls().get(func):
        return await func_caller.run(**arguments)

    if mcp_result := await handle_mcp_tool(func, arguments):
        return mcp_result

    return "(Unknown Function)"



import asyncio
import json
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import (
    AsyncGenerator,
    Generator,
    List,
    Literal,
    Union,
    overload,
)

import dashscope
from dashscope.api_entities.dashscope_response import (
    GenerationResponse,
    MultiModalConversationResponse,
)
from nonebot import logger

from ._types import (
    BasicModel,
    ModelCompletions,
    ModelConfig,
    ModelRequest,
    ModelStreamCompletions,
)
from .utils.tools import function_call_handler


@dataclass
class FunctionCallStream:
    enable: bool = False
    id: str = ""
    function_name: str = ""
    function_args: str = ""

    def from_chunk(self, chunk: GenerationResponse | MultiModalConversationResponse):
        tool_calls = chunk.output.choices[0].message.tool_calls
        tool_call = tool_calls[0]

        if tool_call.get("id", ""):
            self.id = tool_call["id"]

        if tool_call.get("function", {}).get("name", ""):
            self.function_name = tool_call.get("function").get("name")

        function_arg = tool_call.get("function", {}).get("arguments", "")

        if function_arg and self.function_args != function_arg:
            self.function_args += function_arg

        self.enable = True


class ThoughtStream:
    def __init__(self):
        self.is_insert_think_label: bool = False

    def process_chunk(self, chunk: GenerationResponse | MultiModalConversationResponse) -> str:
        choice = chunk.output.choices[0].message
        answer_content = choice.content
        reasoning_content = choice.get("reasoning_content", "")
        reasoning_content = reasoning_content.replace("\n</think>", "") if reasoning_content else ""

        # 处理模型可能输出的 reasoning（思考内容）
        if reasoning_content:
            if not self.is_insert_think_label:
                self.is_insert_think_label = True
                return f"<think>{reasoning_content}"
            else:
                return reasoning_content

        if not answer_content:
            answer_content = ""

        if isinstance(answer_content, list):
            answer_content = answer_content[0].get("text", "")

        if self.is_insert_think_label:
            self.is_insert_think_label = False
            return f"</think>{answer_content}"

        return answer_content


class Dashscope(BasicModel):
    def __init__(self, model_config: ModelConfig) -> None:
        super().__init__(model_config)
        self._require("api_key", "model_name")
        self.api_key = self.config.api_key
        self.model = self.config.model_name
        self.max_tokens = self.config.max_tokens
        self.temperature = self.config.temperature
        self.top_p = self.config.top_p
        self.repetition_penalty = self.config.repetition_penalty
        self.enable_search = self.config.online_search
        self.enable_thinking = self.config.enable_thinking
        self.thinking_budget = self.config.thinking_budget

        self._tools: List[dict] = []
        self._last_call_total_tokens = 0

        self.extra_headers = (
            {"X-DashScope-DataInspection": '{"input":"cip","output":"cip"}'} if self.config.content_security else {}
        )

        self.stream = False

    def __build_multi_messages(self, request: ModelRequest) -> dict:
        """
        构建多模态类型

        此模型加载器支持的多模态类型: `audio` `image`
        """
        multi_contents: List[dict[str, str]] = []

        for item in request.resources:
            if item.type == "audio":
                if not (item.path.startswith("http") or item.path.startswith("file:")):
                    item.url = str(Path(item.path).resolve())

                multi_contents.append({"audio": item.path})

            elif item.type == "image":
                if not (item.path.startswith("http") or item.path.startswith("file:")):
                    item.url = str(Path(item.path).resolve())

                multi_contents.append({"image": item.path})

        user_content = [image_content for image_content in multi_contents]

        if not request.prompt:
            request.prompt = "请描述图像内容"
        user_content.append({"text": request.prompt})

        return {"role": "user", "content": user_content}

    def _build_messages(self, request: ModelRequest) -> list:
        messages = []

        if request.system:
            messages.append({"role": "system", "content": request.system})

        for msg in request.history:
            user_msg = (
                self.__build_multi_messages(ModelRequest(msg.message, resources=msg.resources))
                if all((self.config.multimodal, msg.resources))
                else {"role": "user", "content": msg.message}
            )
            messages.append(user_msg)
            messages.append({"role": "assistant", "content": msg.respond})

        user_msg = (
            {"role": "user", "content": request.prompt}
            if not request.resources
            else self.__build_multi_messages(ModelRequest(request.prompt, resources=request.resources))
        )

        messages.append(user_msg)

        return messages

    async def _GenerationResponse_handle(
        self, messages: list, response: GenerationResponse | MultiModalConversationResponse
    ) -> ModelCompletions:
        completions = ModelCompletions()

        if response.status_code != 200:
            completions.succeed = False
            logger.error(f"模型调用失败: {response.status_code}({response.code})")
            logger.error(f"{response.message}")
            completions.text = f"模型调用失败: {response.status_code}({response.code})"
            return completions

        self._total_tokens += int(response.usage.total_tokens)
        completions.usage = self._total_tokens

        if response.output.text:
            completions.text = response.output.text
            return completions

        message_content = response.output.choices[0].message.content
        if message_content:
            completions.text = message_content if isinstance(message_content, str) else message_content[0].get("text")
            return completions

        return await self._tool_calls_handle_sync(messages, response)

    async def _Generator_handle(
        self,
        messages: list,
        response: Generator[GenerationResponse, None, None] | Generator[MultiModalConversationResponse, None, None],
    ) -> AsyncGenerator[ModelStreamCompletions, None]:
        func_stream = FunctionCallStream()
        thought_stream = ThoughtStream()

        for chunk in response:
            logger.debug(chunk)
            stream_completions = ModelStreamCompletions()

            if chunk.status_code != 200:
                logger.error(f"模型调用失败: {chunk.status_code}({chunk.code})")
                logger.error(f"{chunk.message}")
                stream_completions.chunk = f"模型调用失败: {chunk.status_code}({chunk.code})"
                stream_completions.succeed = False

                yield stream_completions
                return

            # 更新 token 消耗
            current_call_total = chunk.usage.total_tokens
            delta = current_call_total - self._last_call_total_tokens
            if delta < 0:
                delta = current_call_total
            self._total_tokens += delta
            self._last_call_total_tokens = current_call_total
            stream_completions.usage = self._total_tokens

            # 优先判断是否是工具调用（OpenAI-style function calling）
            if chunk.output.choices and chunk.output.choices[0].message.get("tool_calls", []):
                func_stream.from_chunk(chunk)
                # 工具调用也可能在输出文本之后发生

            # DashScope 的 text 模式（非标准接口）
            if hasattr(chunk.output, "text") and chunk.output.text:
                stream_completions.chunk = chunk.output.text
                yield stream_completions
                continue

            if chunk.output.choices is None:
                continue

            stream_completions.chunk = thought_stream.process_chunk(chunk)
            yield stream_completions

        # 流式处理工具调用响应
        if func_stream.enable:
            self._last_call_total_tokens = 0
            async for final_chunk in await self._tool_calls_handle_stream(messages, func_stream):
                yield final_chunk

    async def _tool_calls_handle_sync(
        self, messages: List, response: GenerationResponse | MultiModalConversationResponse
    ) -> ModelCompletions:
        tool_call = response.output.choices[0].message.tool_calls[0]
        tool_call_id = tool_call["id"]
        function_name = tool_call["function"]["name"]
        function_args = json.loads(tool_call["function"]["arguments"])

        logger.info(f"function call 请求 {function_name}, 参数: {function_args}")
        function_return = await function_call_handler(function_name, function_args)
        logger.success(f"Function call 成功，返回: {function_return}")

        messages.append(response.output.choices[0].message)
        messages.append({"role": "tool", "content": function_return, "tool_call_id": tool_call_id})

        return await self._ask(messages)  # type:ignore

    async def _tool_calls_handle_stream(
        self, messages: List, func_stream: FunctionCallStream
    ) -> AsyncGenerator[ModelStreamCompletions, None]:
        function_args = json.loads(func_stream.function_args)

        logger.info(f"function call 请求 {func_stream.function_name}, 参数: {function_args}")
        function_return = await function_call_handler(func_stream.function_name, function_args)  # type:ignore
        logger.success(f"Function call 成功，返回: {function_return}")

        messages.append(
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": func_stream.id,
                        "function": {
                            "arguments": func_stream.function_args,
                            "name": func_stream.function_name,
                        },
                        "type": "function",
                        "index": 0,
                    }
                ],
            }
        )
        messages.append({"role": "tool", "content": function_return, "tool_call_id": func_stream.id})

        return await self._ask(messages)  # type:ignore

    async def _ask(self, messages: list) -> Union[ModelCompletions, AsyncGenerator[ModelStreamCompletions, None]]:
        loop = asyncio.get_event_loop()

        # 因为 Dashscope 对于多模态模型的接口不同，所以这里不能统一函数
        if not self.config.multimodal:
            response = await loop.run_in_executor(
                None,
                partial(
                    dashscope.Generation.call,
                    api_key=self.api_key,
                    model=self.model,
                    messages=messages,
                    max_tokens=self.max_tokens,
                    temperature=self.temperature,
                    top_p=self.top_p,
                    repetition_penalty=self.repetition_penalty,
                    stream=self.stream,
                    tools=self._tools,
                    parallel_tool_calls=True,
                    enable_search=self.enable_search,
                    incremental_output=self.stream,  # 给他调成一样的：这个参数只支持流式调用时设置为True
                    headers=self.extra_headers,
                    enable_thinking=self.enable_thinking,
                    thinking_budget=self.thinking_budget,
                ),
            )
        else:
            response = await loop.run_in_executor(
                None,
                partial(
                    dashscope.MultiModalConversation.call,
                    api_key=self.api_key,
                    model=self.model,
                    messages=messages,
                    max_tokens=self.max_tokens,
                    temperature=self.temperature,
                    top_p=self.top_p,
                    repetition_penalty=self.repetition_penalty,
                    stream=self.stream,
                    tools=self._tools,
                    parallel_tool_calls=True,
                    enable_search=self.enable_search,
                    incremental_output=self.stream,
                ),
            )

        if isinstance(response, GenerationResponse) or isinstance(response, MultiModalConversationResponse):
            return await self._GenerationResponse_handle(messages, response)
        return self._Generator_handle(messages, response)

    @overload
    async def ask(self, request: ModelRequest, *, stream: Literal[False] = False) -> ModelCompletions: ...

    @overload
    async def ask(
        self, request: ModelRequest, *, stream: Literal[True] = True
    ) -> AsyncGenerator[ModelStreamCompletions, None]: ...

    async def ask(
        self, request: ModelRequest, *, stream: bool = False
    ) -> Union[ModelCompletions, AsyncGenerator[ModelStreamCompletions, None]]:
        self._total_tokens = 0
        self._last_call_total_tokens = 0
        self.stream = stream if stream is not None else False

        self._tools = request.tools if request.tools else []
        messages = self._build_messages(request)

        return await self._ask(messages)



import asyncio
from typing import Dict, List, Optional

from nonebot import logger
from nonebot.adapters import Event
from nonebot_plugin_alconna.uniseg import UniMessage, UniMsg

from ..config import plugin_config


class SessionManager:
    def __init__(self) -> None:
        self.sessions: Dict[str, List[UniMsg]] = {}
        self._lock: asyncio.Lock = asyncio.Lock()
        self._timeout = plugin_config.input_timeout

    async def _put(self, sid: str, msg: UniMsg) -> None:
        async with self._lock:
            if sid not in self.sessions:
                self.sessions[sid] = []
            self.sessions[sid].append(msg)

    async def _get_messages_length(self, sid: str) -> int:
        async with self._lock:
            return len(self.sessions.get(sid, []))

    def merge_messages(self, sid: str) -> UniMessage:
        merged_message = UniMessage()

        for message in self.sessions.pop(sid, []):
            merged_message += message

        return merged_message

    async def put_and_wait(self, event: Event, message: UniMsg) -> Optional[UniMessage]:
        sid = event.get_session_id()
        await self._put(sid, message)

        old_length = await self._get_messages_length(sid)
        logger.debug(f"开始等待后续消息 ({self._timeout}s): 会话 {sid}, 当前消息数 {old_length}")
        await asyncio.sleep(self._timeout)
        new_length = await self._get_messages_length(sid)

        if new_length != old_length:
            logger.debug(f"发现新消息插入，当前处理器退出，会话 {sid} 交由后续处理器处理")
            return None

        logger.debug(f"无新消息，当前处理器接管会话 {sid}")
        return self.merge_messages(sid)



from typing import Any, Callable, Coroutine, TypeVar, Union

SYNC_FUNCTION_CALL_FUNC = Callable[..., str]
ASYNC_FUNCTION_CALL_FUNC = Callable[..., Coroutine[str, Any, str]]
FUNCTION_CALL_FUNC = Union[SYNC_FUNCTION_CALL_FUNC, ASYNC_FUNCTION_CALL_FUNC]

F = TypeVar("F", bound=FUNCTION_CALL_FUNC)



from typing import AsyncGenerator, List, Literal, Union, overload

import ollama
from nonebot import logger
from ollama import ResponseError

from ._types import (
    BasicModel,
    ModelCompletions,
    ModelConfig,
    ModelRequest,
    ModelStreamCompletions,
)
from .utils.images import get_file_base64
from .utils.tools import function_call_handler


class Ollama(BasicModel):
    """
    使用 Ollama 模型服务调用模型
    """

    def __init__(self, model_config: ModelConfig) -> None:
        super().__init__(model_config)
        self._require("model_name")
        self.model = self.config.model_name
        self.host = self.config.api_host if self.config.api_host else "http://localhost:11434"
        self.top_k = self.config.top_k
        self.top_p = self.config.top_p
        self.temperature = self.config.temperature
        self.repeat_penalty = self.config.repetition_penalty or 1
        self.presence_penalty = self.config.presence_penalty or 0
        self.frequency_penalty = self.config.frequency_penalty or 1
        self.stream = self.config.stream

        self._tools: List[dict] = []

    def load(self) -> bool:
        try:
            self.client = ollama.AsyncClient(host=self.host)
            self.is_running = True
        except ResponseError as e:
            logger.error(f"加载 Ollama 加载器时发生错误： {e}")
        except ConnectionError as e:
            logger.error(f"加载 Ollama 加载器时发生错误： {e}")
        finally:
            return self.is_running

    def __build_multi_messages(self, request: ModelRequest) -> dict:
        """
        构建多模态类型

        当前模型加载器支持的多模态类型: `image`
        """
        images = []

        for resource in request.resources:
            if resource.path is None:
                continue
            image_base64 = get_file_base64(local_path=resource.path)
            images.append(image_base64)

        message = {"role": "user", "content": request.prompt, "images": images}

        return message

    def _build_messages(self, request: ModelRequest) -> list:
        messages = []

        if request.system:
            messages.append({"role": "system", "content": request.system})

        for index, item in enumerate(request.history):
            messages.append(self.__build_multi_messages(ModelRequest(item.message, resources=item.resources)))
            messages.append({"role": "assistant", "content": item.respond})

        message = self.__build_multi_messages(request)

        messages.append(message)

        return messages

    async def _ask_sync(self, messages: list) -> ModelCompletions:
        completions = ModelCompletions()

        try:
            response = await self.client.chat(
                model=self.model,
                messages=messages,
                tools=self._tools,
                stream=False,
                options={
                    "temperature": self.temperature,
                    "top_k": self.top_k,
                    "top_p": self.top_p,
                    "repeat_penalty": self.repeat_penalty,
                    "presence_penalty": self.presence_penalty,
                    "frequency_penalty": self.frequency_penalty,
                },
            )

            tool_calls = response.message.tool_calls

            if not tool_calls:
                completions.text = response.message.content or "(警告：模型无返回)"
                return completions

            for tool in tool_calls:
                function_name = tool.function.name
                function_args = tool.function.arguments

                logger.info(f"function call 请求 {function_name}, 参数: {function_args}")
                function_return = await function_call_handler(function_name, dict(function_args))
                logger.success(f"Function call 成功，返回: {function_return}")

                messages.append(response.message)
                messages.append({"role": "tool", "content": str(function_return), "name": tool.function.name})
                return await self._ask_sync(messages)

            completions.text = "模型调用错误：未知错误"
            completions.succeed = False
            return completions

        except ollama.ResponseError as e:
            error_info = f"模型调用错误: {e.error}"
            logger.error(error_info)
            completions.succeed = False
            completions.text = error_info
            return completions

    async def _ask_stream(self, messages: list) -> AsyncGenerator[ModelStreamCompletions, None]:
        try:
            response = await self.client.chat(
                model=self.model,
                messages=messages,
                tools=self._tools,
                stream=True,
                options={
                    "temperature": self.temperature,
                    "top_k": self.top_k,
                    "top_p": self.top_p,
                    "repeat_penalty": self.repeat_penalty,
                    "presence_penalty": self.presence_penalty,
                    "frequency_penalty": self.frequency_penalty,
                },
            )

            async for chunk in response:
                stream_completions = ModelStreamCompletions()

                tool_calls = chunk.message.tool_calls

                if chunk.message.content:
                    stream_completions.chunk = chunk.message.content
                    yield stream_completions
                    continue

                if not tool_calls:
                    continue

                for tool in tool_calls:  # type:ignore
                    function_name = tool.function.name
                    function_args = tool.function.arguments

                    logger.info(f"function call 请求 {function_name}, 参数: {function_args}")
                    function_return = await function_call_handler(function_name, dict(function_args))
                    logger.success(f"Function call 成功，返回: {function_return}")

                    messages.append(chunk.message)  # type:ignore
                    messages.append({"role": "tool", "content": str(function_return), "name": tool.function.name})

                    async for content in self._ask_stream(messages):
                        yield content

        except ollama.ResponseError as e:
            stream_completions = ModelStreamCompletions()
            error_info = f"模型调用错误: {e.error}"
            logger.error(error_info)
            stream_completions.chunk = error_info
            stream_completions.succeed = False
            yield stream_completions
            return

    @overload
    async def ask(self, request: ModelRequest, *, stream: Literal[False] = False) -> ModelCompletions: ...

    @overload
    async def ask(
        self, request: ModelRequest, *, stream: Literal[True] = True
    ) -> AsyncGenerator[ModelStreamCompletions, None]: ...

    async def ask(
        self, request: ModelRequest, *, stream: bool = False
    ) -> Union[ModelCompletions, AsyncGenerator[ModelStreamCompletions, None]]:
        self._tools = request.tools if request.tools else []
        messages = self._build_messages(request)

        if stream:
            return self._ask_stream(messages)

        return await self._ask_sync(messages)



import asyncio
import logging
import os
import shutil
from contextlib import AsyncExitStack
from typing import Any, Optional

from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client

from .config import mcpServer


class Tool:
    """
    MCP Tool
    """

    def __init__(self, name: str, description: str, input_schema: dict[str, Any]) -> None:
        self.name: str = name
        self.description: str = description
        self.input_schema: dict[str, Any] = input_schema

    def format_for_llm(self) -> str:
        """
        为 llm 生成工具描述

        :return: 工具描述
        """
        args_desc = []
        if "properties" in self.input_schema:
            for param_name, param_info in self.input_schema["properties"].items():
                arg_desc = f"- {param_name}: {param_info.get('description', 'No description')}"
                if param_name in self.input_schema.get("required", []):
                    arg_desc += " (required)"
                args_desc.append(arg_desc)

        return f"Tool: {self.name}\n" f"Description: {self.description}\n" f"Arguments:{chr(10).join(args_desc)}" ""


class Server:
    """
    管理 MCP 服务器连接和工具执行的 Server 实例
    """

    def __init__(self, name: str, config: mcpServer) -> None:
        self.name: str = name
        self.config: mcpServer = config
        self.stdio_context: Any | None = None
        self.session: ClientSession | None = None
        self._cleanup_lock: asyncio.Lock = asyncio.Lock()
        self.exit_stack: AsyncExitStack = AsyncExitStack()

    async def initialize(self) -> None:
        """
        初始化实例
        """
        command = shutil.which("npx") if self.config.command == "npx" else self.config.command
        if command is None:
            raise ValueError(f"command 字段必须为一个有效值, 且目标指令必须存在于环境变量中: {self.config.command}")

        server_params = StdioServerParameters(
            command=command,
            args=self.config.args,
            env={**os.environ, **self.config.env} if self.config.env else None,
        )
        try:
            stdio_transport = await self.exit_stack.enter_async_context(stdio_client(server_params))
            read, write = stdio_transport
            session = await self.exit_stack.enter_async_context(ClientSession(read, write))
            await session.initialize()
            self.session = session
        except Exception as e:
            logging.error(f"初始化 MCP Server 实例时遇到错误 {self.name}: {e}")
            await self.cleanup()
            raise

    async def list_tools(self) -> list[Tool]:
        """
        从 MCP 服务器获得可用工具列表

        :return: 工具列表

        :raises RuntimeError: 如果服务器未启动
        """
        if not self.session:
            raise RuntimeError(f"Server {self.name} not initialized")

        tools_response = await self.session.list_tools()
        tools: list[Tool] = []

        for item in tools_response:
            if isinstance(item, tuple) and item[0] == "tools":
                tools.extend(Tool(tool.name, tool.description, tool.inputSchema) for tool in item[1])

        return tools

    async def execute_tool(
        self,
        tool_name: str,
        arguments: Optional[dict[str, Any]] = None,
        retries: int = 2,
        delay: float = 1.0,
    ) -> Any:
        """
        执行一个 MCP 工具

        :param tool_name: 工具名称
        :param arguments: 工具参数
        :param retries: 重试次数
        :param delay: 重试间隔

        :return: 工具执行结果

        :raises RuntimeError: 如果服务器未初始化
        :raises Exception: 工具在所有重试中均失败
        """
        if not self.session:
            raise RuntimeError(f"Server {self.name} not initialized")

        attempt = 0
        while attempt < retries:
            try:
                logging.info(f"Executing {tool_name}...")
                result = await self.session.call_tool(tool_name, arguments)

                return result

            except Exception as e:
                attempt += 1
                logging.warning(f"Error executing tool: {e}. Attempt {attempt} of {retries}.")
                if attempt < retries:
                    logging.info(f"Retrying in {delay} seconds...")
                    await asyncio.sleep(delay)
                else:
                    logging.error("Max retries reached. Failing.")
                    raise

    async def cleanup(self) -> None:
        """Clean up server resources."""
        async with self._cleanup_lock:
            try:
                await self.exit_stack.aclose()
                self.session = None
                self.stdio_context = None
            except Exception as e:
                logging.error(f"Error during cleanup of server {self.name}: {e}")



from enum import Enum, auto
from typing import Awaitable, Callable, Union

from muicebot.llm import ModelCompletions, ModelRequest, ModelStreamCompletions
from muicebot.models import Message

SYNC_HOOK_FUNC = Callable[..., None]
ASYNC_HOOK_FUNC = Callable[..., Awaitable[None]]
HOOK_FUNC = Union[SYNC_HOOK_FUNC, ASYNC_HOOK_FUNC]

HOOK_ARGS = Union[Message, ModelCompletions, ModelStreamCompletions, ModelRequest]


class HookType(Enum):
    """可用的 Hook 类型"""

    BEFORE_PRETREATMENT = auto()
    """预处理前"""
    BEFORE_MODEL_COMPLETION = auto()
    """模型调用前"""
    ON_STREAM_CHUNK = auto()
    """模型流式输出中"""
    AFTER_MODEL_COMPLETION = auto()
    """模型调用后"""
    ON_FINISHING_CHAT = auto()
    """结束对话时(存库前)"""



from nonebot import require

require("nonebot_plugin_alconna")
require("nonebot_plugin_localstore")
require("nonebot_plugin_apscheduler")

from nonebot.plugin import PluginMetadata, inherit_supported_adapters  # noqa: E402

from .config import PluginConfig  # noqa: E402
from .utils.utils import init_logger  # noqa: E402

init_logger()

from . import onebot  # noqa: E402, F401

__plugin_meta__ = PluginMetadata(
    name="MuiceBot",
    description="Muice-Chatbot 的 Nonebot2 实现，支持市面上大多数的模型",
    usage="@at / {config.MUICE_NICKNAMES} <message>: 与大语言模型交互；关于指令类可输入 .help 查询",
    type="application",
    config=PluginConfig,
    homepage="https://bot.snowy.moe/",
    extra={},
    supported_adapters=inherit_supported_adapters("nonebot_plugin_alconna"),
)



import re
from dataclasses import dataclass

from nonebot.adapters import Event

from muicebot.config import plugin_config
from muicebot.llm import ModelCompletions, ModelStreamCompletions
from muicebot.models import Message
from muicebot.plugin.hook import on_after_completion, on_finish_chat, on_stream_chunk

_PROCESS_MODE = plugin_config.thought_process_mode
_STREAM_PROCESS_STATE: dict[str, bool] = {}
_PROCESSCACHES: dict[str, "ProcessCache"] = {}


@dataclass
class ProcessCache:
    thoughts: str = ""
    result: str = ""


def general_processor(message: str) -> tuple[str, str]:
    thoughts_pattern = re.compile(r"<think>(.*?)</think>", re.DOTALL)
    match = thoughts_pattern.search(message)
    thoughts = match.group(1).replace("\n", "") if match else ""
    result = thoughts_pattern.sub("", message).strip()
    return thoughts, result


@on_after_completion(priority=1, stream=False)
def async_processor(completions: ModelCompletions, event: Event):
    session_id = event.get_session_id()
    thoughts, result = general_processor(completions.text)
    _PROCESSCACHES[session_id] = ProcessCache(thoughts, result)

    if _PROCESS_MODE == 0:
        return
    if _PROCESS_MODE == 2 or not thoughts:
        completions.text = result
    elif _PROCESS_MODE == 1:
        completions.text = f"思考过程: {thoughts}\n\n{result}"


@on_stream_chunk(priority=1)
def stream_processor(chunk: ModelStreamCompletions, event: Event):
    session_id = event.get_session_id()
    cache = _PROCESSCACHES.setdefault(session_id, ProcessCache())
    state = _STREAM_PROCESS_STATE

    # 思考过程中
    if "<think>" in chunk.chunk:
        state[session_id] = True
        cache.thoughts += chunk.chunk.replace("<think>", "")
        if _PROCESS_MODE == 1:
            chunk.chunk = chunk.chunk.replace("<think>", "思考过程: ")
        elif _PROCESS_MODE == 2:
            chunk.chunk = ""
        return

    # 思考结束
    elif "</think>" in chunk.chunk:
        del state[session_id]
        cache.result += chunk.chunk.replace("</think>", "")
        if _PROCESS_MODE == 1:
            chunk.chunk = chunk.chunk.replace("</think>", "\n\n")
        elif _PROCESS_MODE == 2:
            chunk.chunk = chunk.chunk.replace("</think>", "")
        return

    # 思考过程中
    elif state.get(session_id, False):
        cache.thoughts += chunk.chunk
        if _PROCESS_MODE == 2:
            chunk.chunk = ""

    # 思考结果中
    else:
        cache.result += chunk.chunk


@on_finish_chat(priority=1)
def save_processor(message: Message, event: Event):
    session_id = event.get_session_id()
    cache = _PROCESSCACHES.pop(session_id, ProcessCache())
    message.respond = cache.result



from dataclasses import asdict, dataclass, field
from datetime import datetime
from functools import total_ordering
from io import BytesIO
from typing import List, Literal, Optional, Union


@dataclass
class Resource:
    """多模态消息"""

    type: Literal["image", "video", "audio", "file"]
    """消息类型"""
    path: str = field(default_factory=str)
    """本地存储地址(对于模型处理是必需的)"""
    url: Optional[str] = field(default=None)
    """远程存储地址(一般不传入模型处理中)"""
    raw: Optional[Union[bytes, BytesIO]] = field(default=None)
    """二进制数据（只使用于模型返回且不保存到数据库中）"""
    mimetype: Optional[str] = field(default=None)
    """文件元数据类型"""

    def ensure_mimetype(self):
        from .utils.utils import guess_mimetype

        if not self.mimetype:
            self.mimetype = guess_mimetype(self)

    def to_dict(self) -> dict:
        return {
            "type": self.type,
            "path": self.path,
        }


@total_ordering
@dataclass
class Message:
    """格式化后的 bot 消息"""

    id: int | None = None
    """每条消息的唯一ID"""
    time: str = field(default_factory=lambda: datetime.strftime(datetime.now(), "%Y.%m.%d %H:%M:%S"))
    """
    字符串形式的时间数据：%Y.%m.%d %H:%M:%S
    若要获取格式化的 datetime 对象，请使用 format_time
    """
    userid: str = ""
    """Nonebot 的用户id"""
    groupid: str = "-1"
    """群组id，私聊设为-1"""
    message: str = ""
    """消息主体"""
    respond: str = ""
    """模型回复（不包含思维过程）"""
    history: int = 1
    """消息是否可用于对话历史中，以整数形式映射布尔值"""
    resources: List[Resource] = field(default_factory=list)
    """多模态消息内容"""
    usage: int = -1
    """使用的总 tokens, 若模型加载器不支持则设为-1"""

    @property
    def format_time(self) -> datetime:
        """将时间字符串转换为 datetime 对象"""
        return datetime.strptime(self.time, "%Y.%m.%d %H:%M:%S")

    def to_dict(self) -> dict:
        return asdict(self)

    @staticmethod
    def from_dict(data: dict) -> "Message":
        return Message(**data)

    def __hash__(self) -> int:
        return hash(self.id)

    def __lt__(self, other: "Message") -> bool:
        return self.format_time < other.format_time



import atexit
import os
import threading
import time
from pathlib import Path
from typing import Callable, List, Literal, Optional

import yaml as yaml_
from nonebot import get_driver, get_plugin_config, logger
from nonebot.config import Config
from pydantic import BaseModel
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from .llm import ModelConfig

MODELS_CONFIG_PATH = Path("configs/models.yml").resolve()
SCHEDULES_CONFIG_PATH = Path("configs/schedules.yml").resolve()
PLUGINS_CONFIG_PATH = Path("configs/plugins.yml").resolve()


def load_yaml_config() -> dict:
    """
    插件优先加载 YAML 配置，失败则返回空字典
    """
    try:
        with open(PLUGINS_CONFIG_PATH, "r", encoding="utf-8") as f:
            return yaml_.safe_load(f) or {}
    except (FileNotFoundError, yaml_.YAMLError):
        return {}


driver = get_driver()
yaml_config = load_yaml_config()
env_config = driver.config.model_dump()
final_config = {**env_config, **yaml_config}  # 合并配置，yaml优先
driver.config = Config(**final_config)


class PluginConfig(BaseModel):
    log_level: str = "INFO"
    """日志等级"""
    muice_nicknames: list = ["muice"]
    """沐雪的自定义昵称，作为消息前缀条件响应信息事件"""
    telegram_proxy: str | None = None
    """telegram代理，这个配置项用于获取图片时使用"""
    enable_builtin_plugins: bool = True
    """启用内嵌插件"""
    max_history_epoch: int = 0
    """最大历史轮数"""
    enable_adapters: list = ["nonebot.adapters.onebot.v11", "nonebot.adapters.onebot.v12"]
    """启用的 Nonebot 适配器"""
    input_timeout: int = 0
    """输入等待时间"""
    default_template: Optional[str] = None
    """默认使用人设模板名称"""
    thought_process_mode: Literal[0, 1, 2] = 2
    """针对 Deepseek-R1 等思考模型的思考过程提取模式"""


plugin_config = get_plugin_config(PluginConfig)


class Schedule(BaseModel):
    id: str
    """调度器 ID"""
    trigger: Literal["cron", "interval"]
    """调度器类别"""
    ask: Optional[str] = None
    """向大语言模型询问的信息"""
    say: Optional[str] = None
    """直接输出的信息"""
    args: dict[str, int]
    """调度器参数"""
    target: str
    """目标id；若为群聊则为 group_id 或者 channel_id，若为私聊则为 user_id"""
    probability: int = 1
    """触发几率"""


def get_schedule_configs() -> List[Schedule]:
    """
    从配置文件 `configs/schedules.yml` 中获取所有调度器配置

    如果没有该文件，返回空列表
    """
    if not os.path.isfile(SCHEDULES_CONFIG_PATH):
        return []

    with open(SCHEDULES_CONFIG_PATH, "r", encoding="utf-8") as f:
        configs = yaml_.safe_load(f)

    if not configs:
        return []

    schedule_configs = []

    for schedule_id, config in configs.items():
        config["id"] = schedule_id
        schedule_config = Schedule(**config)
        schedule_configs.append(schedule_config)

    return schedule_configs


class ConfigFileHandler(FileSystemEventHandler):
    """配置文件变化处理器"""

    def __init__(self, callback: Callable):
        self.callback = callback
        self.last_modified = time.time()
        # 防止一次修改触发多次回调
        self.cooldown = 1  # 冷却时间（秒）

    def on_modified(self, event):
        if event.is_directory:  # 检查是否是文件而不是目录
            return

        current_time = time.time()
        if current_time - self.last_modified > self.cooldown:
            self.last_modified = current_time
            self.callback()


class ModelConfigManager:
    """模型配置管理器"""

    _instance: Optional["ModelConfigManager"] = None
    _lock = threading.Lock()
    _initialized: bool
    configs: dict[str, ModelConfig]

    def __new__(cls):
        """确保实例在单例模式下运行"""
        with cls._lock:
            if cls._instance is None:
                cls._instance = super(ModelConfigManager, cls).__new__(cls)
                cls._instance._initialized = False
            return cls._instance

    def __init__(self) -> None:
        if self._initialized:
            return

        self.configs = {}
        self.default_config = None
        self.observer = None
        self.listeners: List[Callable] = []  # 注册的监听器列表
        self._load_configs()
        self._start_file_watcher()
        self._initialized = True

    def _load_configs(self):
        """加载配置文件"""
        if not os.path.isfile(MODELS_CONFIG_PATH):
            raise FileNotFoundError("configs/models.yml 不存在！请先创建")

        with open(MODELS_CONFIG_PATH, "r", encoding="utf-8") as f:
            configs_dict = yaml_.safe_load(f)

        if not configs_dict:
            raise ValueError("configs/models.yml 为空，请先至少定义一个模型配置")

        self.configs = {}
        for name, config in configs_dict.items():
            self.configs[name] = ModelConfig(**config)
            # 未指定模板时，使用默认模板
            self.configs[name].template = self.configs[name].template or plugin_config.default_template
            if config.get("default"):
                self.default_config = self.configs[name]

        if not self.default_config and self.configs:
            # 如果没有指定默认配置，使用第一个
            self.default_config = next(iter(self.configs.values()))

    def _start_file_watcher(self):
        """启动文件监视器"""
        if self.observer is not None:
            self.observer.stop()

        self.observer = Observer()
        event_handler = ConfigFileHandler(self._on_config_changed)
        self.observer.schedule(event_handler, str(Path(MODELS_CONFIG_PATH).parent), recursive=False)
        self.observer.start()

    def _on_config_changed(self):
        """配置文件变化时的回调函数"""
        try:
            # old_configs = self.configs.copy()
            old_default = self.default_config

            self._load_configs()

            # 通知所有注册的监听器
            for listener in self.listeners:
                listener(self.default_config, old_default)

        except Exception as e:
            logger.error(f"重新加载配置文件失败: {e}")

    def register_listener(self, listener: Callable):
        """
        注册配置变化监听器

        :param listener: 回调函数，接收两个参数：新的默认配置和旧的默认配置
        """
        if listener not in self.listeners:
            self.listeners.append(listener)

    def unregister_listener(self, listener: Callable):
        """取消注册配置变化监听器"""
        if listener in self.listeners:
            self.listeners.remove(listener)

    def get_model_config(self, model_config_name: Optional[str] = None) -> ModelConfig:
        """获取指定模型的配置"""
        if model_config_name in [None, ""]:
            if not self.default_config:
                raise ValueError("没有找到默认模型配置！请确保存在至少一个有效的配置项！")
            return self.default_config

        elif model_config_name in self.configs:
            return self.configs[model_config_name]

        else:
            logger.warning(f"指定的模型配置 '{model_config_name}' 不存在！")
            raise ValueError(f"指定的模型配置 '{model_config_name}' 不存在！")

    def stop_watcher(self):
        """停止文件监视器"""
        if self.observer is None:
            return

        self.observer.stop()
        self.observer.join()


model_config_manager = ModelConfigManager()
atexit.register(model_config_manager.stop_watcher)


def get_model_config(model_config_name: Optional[str] = None) -> ModelConfig:
    """
    从配置文件 `configs/models.yml` 中获取指定模型的配置文件

    :model_config_name: (可选)模型配置名称。若为空，则先寻找配置了 `default: true` 的首个配置项，若失败就再寻找首个配置项
    若都不存在，则抛出 `FileNotFoundError`
    """
    return model_config_manager.get_model_config(model_config_name)



from functools import wraps
from typing import Any

from ._types import ASYNC_FUNCTION_CALL_FUNC, SYNC_FUNCTION_CALL_FUNC


def async_wrap(func: SYNC_FUNCTION_CALL_FUNC) -> ASYNC_FUNCTION_CALL_FUNC:
    """
    装饰器，将同步函数包装为异步函数
    """

    @wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        return func(*args, **kwargs)

    return wrapper



import asyncio
import json
import os
from datetime import datetime
from typing import Optional, Tuple

import aiosqlite
import nonebot_plugin_localstore as store
from nonebot import logger

from .models import Message, Resource
from .utils.migrations import MigrationManager


class Database:
    def __init__(self) -> None:
        self.DB_PATH = store.get_plugin_data_dir().joinpath("ChatHistory.db").resolve()
        self.migrations = MigrationManager(self)

        asyncio.run(self.init_db())

        logger.info(f"数据库路径: {self.DB_PATH}")

    async def init_db(self) -> None:
        """初始化数据库，检查数据库是否存在，不存在则创建"""
        if not os.path.isfile(self.DB_PATH) or self.DB_PATH.stat().st_size == 0:
            logger.info("数据库不存在，正在创建...")
            await self.__create_database()

        await self.migrations.migrate_if_needed()

    def __connect(self) -> aiosqlite.Connection:
        return aiosqlite.connect(self.DB_PATH)

    async def execute(self, query: str, params=(), fetchone=False, fetchall=False) -> list | None:
        """
        异步执行SQL查询，支持可选参数。

        :param query: 要执行的SQL查询语句
        :param params: 传递给查询的参数
        :param fetchone: 是否获取单个结果
        :param fetchall: 是否获取所有结果
        """
        async with self.__connect() as conn:
            conn.row_factory = aiosqlite.Row
            cursor = await conn.cursor()
            await cursor.execute(query, params)
            if fetchone:
                return await cursor.fetchone()  # type: ignore
            if fetchall:
                rows = await cursor.fetchall()
                return [{k.lower(): v for k, v in zip(row.keys(), row)} for row in rows]
            await conn.commit()

        return None

    async def __create_database(self) -> None:
        """
        创建一个新的信息表
        """
        await self.execute(
            """CREATE TABLE MSG(
            ID INTEGER PRIMARY KEY AUTOINCREMENT,
            TIME TEXT NOT NULL,
            USERID TEXT NOT NULL,
            GROUPID TEXT NOT NULL DEFAULT (-1),
            MESSAGE TEXT NOT NULL,
            RESPOND TEXT NOT NULL,
            HISTORY INTEGER NOT NULL DEFAULT (1),
            RESOURCES TEXT NOT NULL DEFAULT "[]",
            USAGE INTEGER NOT NULL DEFAULT (-1));"""
        )
        await self.execute(
            """
            CREATE TABLE schema_version (
                version INTEGER NOT NULL
            );"""
        )
        await self.execute("INSERT INTO schema_version (version) VALUES (?);", (str(self.migrations.latest_version)))

    def connect(self) -> aiosqlite.Connection:
        return aiosqlite.connect(self.DB_PATH)

    async def add_item(self, message: Message):
        """
        将消息保存到数据库
        """
        resources_data = [r.to_dict() for r in message.resources]
        params = (
            message.time,
            message.userid,
            message.groupid,
            message.message,
            message.respond,
            json.dumps(resources_data, ensure_ascii=False),
            message.usage,
        )
        query = """INSERT INTO MSG (TIME, USERID, GROUPID, MESSAGE, RESPOND, RESOURCES, USAGE)
                   VALUES (?, ?, ?, ?, ?, ?, ?)"""
        await self.execute(query, params)

    async def mark_history_as_unavailable(self, userid: str, limit: Optional[int] = None):
        """
        将用户对话历史标记为不可用 (适用于 reset 命令)

        :param userid: 用户id
        :param limit: (可选)操作数量
        """
        if limit is not None:
            query = """UPDATE MSG SET HISTORY = 0 WHERE ROWID IN (
                SELECT ROWID FROM MSG WHERE USERID = ? AND HISTORY = 1 ORDER BY ROWID DESC LIMIT ?)"""
            await self.execute(query, (userid, limit))
        else:
            query = "UPDATE MSG SET HISTORY = 0 WHERE USERID = ?"
            await self.execute(query, (userid,))

    async def _deserialize_rows(self, rows: list) -> list[Message]:
        """
        反序列化数据库返回结果
        """
        result = []

        for row in rows:
            data = dict(row)

            # 反序列化 resources
            resources = json.loads(data.get("resources", "[]"))
            data["resources"] = [Resource(**r) for r in resources] if resources else []

            result.append(Message(**data))

        result.reverse()
        return result

    async def get_user_history(self, userid: str, limit: int = 0) -> list[Message]:
        """
        获取用户的所有对话历史，返回一个列表，无结果时返回None

        :param userid: 用户名
        :limit: (可选) 返回的最大长度，当该变量设为0时表示全部返回
        """
        if limit:
            query = f"SELECT * FROM MSG WHERE HISTORY = 1 AND USERID = ? ORDER BY ID DESC LIMIT {limit}"
        else:
            query = "SELECT * FROM MSG WHERE HISTORY = 1 AND USERID = ?"
        rows = await self.execute(query, (userid,), fetchall=True)

        result = await self._deserialize_rows(rows) if rows else []

        return result

    async def get_group_history(self, groupid: str, limit: int = 0) -> list[Message]:
        """
        获取群组的所有对话历史，返回一个列表，无结果时返回None

        :groupid: 群组id
        :limit: (可选) 返回的最大长度，当该变量设为0时表示全部返回
        """
        if limit:
            query = f"SELECT * FROM MSG WHERE HISTORY = 1 AND GROUPID = ? ORDER BY ID DESC LIMIT {limit}"
        else:
            query = "SELECT * FROM MSG WHERE HISTORY = 1 AND GROUPID = ?"
        rows = await self.execute(query, (groupid,), fetchall=True)

        result = await self._deserialize_rows(rows) if rows else []

        return result

    async def get_model_usage(self) -> Tuple[int, int]:
        """
        获取模型用量数据（今日用量，总用量）

        :return: today_usage, total_usage
        """
        today_str = datetime.now().strftime("%Y.%m.%d")

        # 查询总用量（排除 USAGE = -1）
        total_result = await self.execute("SELECT SUM(USAGE) FROM MSG WHERE USAGE != -1", fetchone=True)
        total_usage = total_result[0] if total_result and total_result[0] is not None else 0

        # 查询今日用量（按日期前缀匹配 TIME）
        today_result = await self.execute(
            "SELECT SUM(USAGE) FROM MSG WHERE USAGE != -1 AND TIME LIKE ?",
            (f"{today_str}%",),
            fetchone=True,
        )
        today_usage = today_result[0] if today_result and today_result[0] is not None else 0

        return today_usage, total_usage

    async def get_conv_count(self) -> Tuple[int, int]:
        """
        获取对话次数（今日次数，总次数）

        :return: today_count, total_count
        """
        today_str = datetime.now().strftime("%Y.%m.%d")

        total_result = await self.execute("SELECT COUNT(*) FROM MSG WHERE USAGE != -1", fetchone=True)
        total_count = total_result[0] if total_result and total_result[0] is not None else 0

        today_result = await self.execute(
            "SELECT COUNT(*) FROM MSG WHERE USAGE != -1 AND TIME LIKE ?",
            (f"{today_str}%",),
            fetchone=True,
        )
        today_count = today_result[0] if today_result and today_result[0] is not None else 0

        return today_count, total_count

    async def remove_last_item(self, userid: str):
        """
        删除用户的最新一条对话历史

        :userid: 用户id
        """
        query = "DELETE FROM MSG WHERE ID = (SELECT ID FROM MSG WHERE USERID = ? ORDER BY ID DESC LIMIT 1)"
        await self.execute(query, (userid,))



import inspect
from collections import defaultdict
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from nonebot import logger
from nonebot.adapters import Bot, Event
from nonebot.matcher import Matcher
from nonebot.rule import Rule
from nonebot.typing import T_State

from ..context import get_bot, get_event, get_mather
from ._types import HOOK_ARGS, HOOK_FUNC, HookType

DEPENDENCY_PROVIDERS: dict[type, Callable[[], object]] = {
    Bot: get_bot,
    Event: get_event,
    Matcher: get_mather,
    # T_State: get_state,
}


def _match_union(param_type: type, arg: object) -> bool:
    if get_origin(param_type) is Union:
        return any(isinstance(arg, t) for t in get_args(param_type))
    return False


class HookManager:
    def __init__(self):
        self._hooks: Dict[HookType, List["Hooked"]] = defaultdict(list)

    async def _inject_dependencies(self, function: HOOK_FUNC, hook_arg: HOOK_ARGS) -> dict:
        """
        自动解析参数并进行依赖注入
        """
        sig = inspect.signature(function)
        hints = get_type_hints(function)

        inject_args: dict[str, Any] = {}

        for name, param in sig.parameters.items():
            param_type = hints.get(name, None)
            if not param_type:
                continue

            # 1. Union 类型注入 hook_arg
            if _match_union(param_type, hook_arg):
                inject_args[name] = hook_arg
                continue

            # 2. 直接匹配 hook_arg
            if isinstance(hook_arg, param_type):
                inject_args[name] = hook_arg
                continue

            # 3. 依赖提供者匹配（Bot、Event、Matcher...）
            for dep_type, provider in DEPENDENCY_PROVIDERS.items():
                if isinstance(param_type, type) and issubclass(param_type, dep_type):
                    inject_args[name] = provider()
                    break

        return inject_args

    def register(self, hook_type: HookType, hooked: "Hooked"):
        """
        注册一个挂钩函数
        """
        self._hooks[hook_type].append(hooked)
        return hooked

    async def run(self, hook_type: HookType, hook_arg: HOOK_ARGS, stream: bool = False):
        """
        运行所有的钩子函数

        :param hook_type: 钩子类型
        :param hook_arg: 消息处理流程中对应的数据类
        :param stream: 当前是否为流式状态
        """
        hookeds = self._hooks[hook_type]
        hookeds.sort(key=lambda x: x.priority)

        bot: Bot = get_bot()
        event: Event = get_event()
        state: T_State = {}

        for hooked in hookeds:
            args = await self._inject_dependencies(hooked.function, hook_arg)

            if (hooked.stream is not None and hooked.stream == stream) or (
                hooked.rule and not await hooked.rule(bot, event, state)
            ):
                continue

            result = hooked.function(**args)
            if isinstance(result, Awaitable):
                await result


hook_manager = HookManager()


class Hooked:
    """挂钩函数对象"""

    def __init__(
        self, hook_type: HookType, priority: int = 10, stream: Optional[bool] = None, rule: Optional[Rule] = None
    ):
        self.hook_type = hook_type
        """钩子函数类型"""
        self.priority = priority
        """函数调用优先级"""
        self.stream = stream
        """是否仅在(非)流式中运行"""
        self.rule: Optional[Rule] = rule
        """启用规则"""

        self.function: HOOK_FUNC
        """函数对象"""

    def __call__(self, func: HOOK_FUNC) -> HOOK_FUNC:
        """
        修饰器：注册一个 Hook 函数
        """
        self.function = func

        # 获取模块名
        if module := inspect.getmodule(func):
            module_name = module.__name__.split(".")[-1]
        else:
            module_name = ""

        hook_manager.register(self.hook_type, self)
        logger.success(f"挂钩函数 {module_name}.{func.__name__} 已成功加载")
        return func


def on_before_pretreatment(priority: int = 10, rule: Optional[Rule] = None) -> Hooked:
    """
    注册一个钩子函数
    这个函数将在传入消息 (`Muice` 的 `_prepare_prompt()`) 前调用
    它可接受一个 `Message` 类参数

    :param priority: 调用优先级
    :param rule: Nonebot 的响应规则
    """
    return Hooked(HookType.BEFORE_PRETREATMENT, priority=priority, rule=rule)


def on_before_completion(priority: int = 10, rule: Optional[Rule] = None) -> Hooked:
    """
    注册一个钩子函数。
    这个函数将在传入模型(`Muice` 的 `model.ask()`)前调用
    它可接受一个 `ModelRequest` 类参数

    :param priority: 调用优先级
    :param rule: Nonebot 的响应规则
    """
    return Hooked(HookType.BEFORE_MODEL_COMPLETION, priority=priority, rule=rule)


def on_stream_chunk(priority: int = 10, rule: Optional[Rule] = None) -> Hooked:
    """
    注册一个钩子函数。
    这个函数将在流式调用中途(`Muice` 的 `model.ask()`)调用
    它可接受一个 `ModelStreamCompletions` 类参数

    :param priority: 调用优先级
    :param rule: Nonebot 的响应规则
    """
    return Hooked(HookType.ON_STREAM_CHUNK, priority=priority, rule=rule)


def on_after_completion(priority: int = 10, stream: Optional[bool] = None, rule: Optional[Rule] = None) -> Hooked:
    """
    注册一个钩子函数。
    这个函数将在传入模型(`Muice` 的 `model.ask()`)后调用（流式则传入整合后的数据）
    它可接受一个 `ModelCompletion` 类参数

    请注意：当启用流式时，对 `ModelStreamCompletion` 的任何修改将不生效

    :param priority: 调用优先级
    :param stream: 是否仅在(非)流式中处理，None 则无限制
    :param rule: Nonebot 的响应规则
    """
    return Hooked(HookType.AFTER_MODEL_COMPLETION, priority=priority, stream=stream, rule=rule)


def on_finish_chat(priority: int = 10, rule: Optional[Rule] = None) -> Hooked:
    """
    注册一个钩子函数。
    这个函数将在结束对话(存库前)调用
    它可接受一个 `Message` 类参数

    :param priority: 调用优先级
    :param rule: Nonebot 的响应规则
    """
    return Hooked(HookType.ON_FINISHING_CHAT, priority=priority, rule=rule)



import importlib
import os
import time
from typing import AsyncGenerator, Optional, Union

from nonebot import logger

from .config import ModelConfig, get_model_config, model_config_manager, plugin_config
from .database import Database
from .llm import (
    MODEL_DEPENDENCY_MAP,
    BasicModel,
    ModelCompletions,
    ModelRequest,
    ModelStreamCompletions,
    get_missing_dependencies,
)
from .models import Message, Resource
from .plugin.func_call import get_function_list
from .plugin.hook import HookType, hook_manager
from .plugin.mcp import get_mcp_list
from .templates import generate_prompt_from_template
from .utils.utils import get_username


class Muice:
    """
    Muice交互类
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        if self._initialized:
            return

        self.model_config = get_model_config()

        self.database = Database()
        self.max_history_epoch = plugin_config.max_history_epoch

        self.system_prompt = ""
        self.user_instructions = ""

        self._load_config()
        self._init_model()

        model_config_manager.register_listener(self._on_config_changed)

        self._initialized = True

    def __del__(self):
        # 注销监听器
        try:
            model_config_manager.unregister_listener(self._on_config_changed)
        except (AttributeError, RuntimeError) as e:
            logger.debug(f"Muice __del__ 清理失败: {e}")

    def _load_config(self):
        """
        加载配置项
        """
        self.model_loader = self.model_config.loader
        self.multimodal = self.model_config.multimodal
        self.template = self.model_config.template
        self.template_mode = self.model_config.template_mode

    def _init_model(self) -> None:
        """
        初始化模型类
        """
        try:
            module_name = f"muicebot.llm.{self.model_loader}"
            module = importlib.import_module(module_name)
            ModelClass = getattr(module, self.model_loader, None)
            self.model: Optional[BasicModel] = ModelClass(self.model_config) if ModelClass else None

        except ImportError as e:
            logger.critical(f"导入模型加载器 '{self.model_loader}' 失败：{e}")
            dependencies = MODEL_DEPENDENCY_MAP.get(self.model_loader, [])
            missing = get_missing_dependencies(dependencies)
            if missing:
                install_command = "pip install " + " ".join(missing)
                logger.critical(f"缺少依赖库：{', '.join(missing)}\n请运行以下命令安装缺失项：\n\n{install_command}")

        except AttributeError as e:
            logger.critical(f"导入模型加载器 '{self.model_loader}' 失败：{e}")
            logger.critical("这有可能是负责编写模型加载器的开发者未正确命名类导致，又或者是您输入了错误的模型加载器名")

    def load_model(self) -> bool:
        """
        加载模型

        return: 是否加载成功
        """
        if not self.model:
            logger.error("模型加载失败: self.model 变量不存在")
            return False
        if not self.model.load():
            logger.error("模型加载失败: self.model.load 函数失败")
            return False
        return True

    def _on_config_changed(self, new_config: ModelConfig, old_config: ModelConfig):
        """配置文件变更时的回调函数"""
        logger.info("检测到配置文件变更，自动重载模型...")
        # 更新配置
        self.model_config = new_config

        # 重新加载模型
        self._load_config()
        self._init_model()
        self.load_model()
        logger.success(f"模型自动重载完成: {old_config.loader} -> {new_config.loader}")

    def change_model_config(self, config_name: str) -> str:
        """
        更换模型配置文件并重新加载模型
        """
        try:
            self.model_config = get_model_config(config_name)
        except (ValueError, FileNotFoundError) as e:
            return str(e)

        self._load_config()
        self._init_model()
        self.load_model()

        return f"已成功加载 {config_name}" if config_name else "未指定模型配置名，已加载默认模型配置"

    async def _prepare_prompt(self, message: str, userid: str, is_private: bool) -> str:
        """
        准备提示词(包含系统提示)

        :param message: 消息主体
        :param userid: 用户 Nonebot ID
        :param is_private: 是否为私聊信息
        :return: 最终模型提示词
        """
        if self.template is None:
            return message

        system_prompt = generate_prompt_from_template(self.template, userid, is_private).strip()

        if self.template_mode == "system":
            self.system_prompt = system_prompt
        else:
            self.user_instructions = system_prompt

        if is_private:
            return f"{self.user_instructions}\n\n{message}" if self.user_instructions else message

        group_prompt = f"<{await get_username()}> {message}"

        return f"{self.user_instructions}\n\n{group_prompt}" if self.user_instructions else group_prompt

    async def _prepare_history(self, userid: str, groupid: str = "-1", enable_history: bool = True) -> list[Message]:
        """
        准备对话历史

        :param userid: 用户名
        :param groupid: 群组ID等(私聊时此值为-1)
        :param enable_history: 是否启用历史记录
        :return: 最终模型提示词
        """
        user_history = await self.database.get_user_history(userid, self.max_history_epoch) if enable_history else []

        # 验证多模态资源路径是否可用
        for item in user_history:
            item.resources = [
                resource for resource in item.resources if resource.path and os.path.isfile(resource.path)
            ]

        if groupid == "-1":
            return user_history[-self.max_history_epoch :]

        group_history = await self.database.get_group_history(groupid, self.max_history_epoch)

        for item in group_history:
            item.resources = [
                resource for resource in item.resources if resource.path and os.path.isfile(resource.path)
            ]

        # 群聊历史构建成 <Username> Message 的格式，避免上下文混乱
        for item in group_history:
            user_name = await get_username(item.userid)
            item.message = f"<{user_name}> {item.message}"

        final_history = list(set(user_history + group_history))

        return final_history[-self.max_history_epoch :]

    async def ask(
        self,
        message: Message,
        enable_history: bool = True,
        enable_plugins: bool = True,
    ) -> ModelCompletions:
        """
        调用模型

        :param message: 消息文本
        :param enable_history: 是否启用历史记录
        :param enable_plugins: 是否启用工具插件
        :return: 模型回复
        """
        if not (self.model and self.model.is_running):
            logger.error("模型未加载")
            return ModelCompletions("模型未加载", succeed=False)

        is_private = message.groupid == "-1"
        logger.info("正在调用模型...")

        await hook_manager.run(HookType.BEFORE_PRETREATMENT, message)

        prompt = await self._prepare_prompt(message.message, message.userid, is_private)
        history = await self._prepare_history(message.userid, message.groupid, enable_history) if enable_history else []
        tools = (
            (await get_function_list() + await get_mcp_list())
            if self.model_config.function_call and enable_plugins
            else []
        )
        system = self.system_prompt if self.system_prompt else None

        model_request = ModelRequest(prompt, history, message.resources, tools, system)
        await hook_manager.run(HookType.BEFORE_MODEL_COMPLETION, model_request)

        start_time = time.perf_counter()
        logger.debug(f"模型调用参数：Prompt: {message}, History: {history}")

        response = await self.model.ask(model_request, stream=False)

        end_time = time.perf_counter()

        logger.success(f"模型调用{'成功' if response.succeed else '失败'}: {response}")
        logger.debug(f"模型调用时长: {end_time - start_time} s (token用量: {response.usage})")

        await hook_manager.run(HookType.AFTER_MODEL_COMPLETION, response)

        message.respond = response.text
        message.usage = response.usage

        await hook_manager.run(HookType.ON_FINISHING_CHAT, message)

        if response.succeed:
            await self.database.add_item(message)

        return response

    async def ask_stream(
        self,
        message: Message,
        enable_history: bool = True,
        enable_plugins: bool = True,
    ) -> AsyncGenerator[ModelStreamCompletions, None]:
        """
        调用模型

        :param message: 消息文本
        :param enable_history: 是否启用历史记录
        :param enable_plugins: 是否启用工具插件
        :return: 模型回复
        """
        if not (self.model and self.model.is_running):
            logger.error("模型未加载")
            yield ModelStreamCompletions("模型未加载")
            return

        is_private = message.groupid == "-1"
        logger.info("正在调用模型...")

        await hook_manager.run(HookType.BEFORE_PRETREATMENT, message)

        prompt = await self._prepare_prompt(message.message, message.userid, is_private)
        history = await self._prepare_history(message.userid, message.groupid, enable_history) if enable_history else []
        tools = (
            (await get_function_list() + await get_mcp_list())
            if self.model_config.function_call and enable_plugins
            else []
        )
        system = self.system_prompt if self.system_prompt else None

        model_request = ModelRequest(prompt, history, message.resources, tools, system)
        await hook_manager.run(HookType.BEFORE_MODEL_COMPLETION, model_request)

        start_time = time.perf_counter()
        logger.debug(f"模型调用参数：Prompt: {message}, History: {history}")

        response = await self.model.ask(model_request, stream=True)

        total_reply = ""
        total_resources: list[Resource] = []
        item: Optional[ModelStreamCompletions] = None

        async for item in response:
            await hook_manager.run(HookType.ON_STREAM_CHUNK, item)
            total_reply += item.chunk
            yield item
            if item.resources:
                total_resources.extend(item.resources)

        if item is None:
            raise RuntimeError("模型调用器返回的值应至少包含一个元素")

        end_time = time.perf_counter()
        logger.success(f"已完成流式回复: {total_reply}")
        logger.debug(f"模型调用时长: {end_time - start_time} s (token用量: {item.usage})")

        final_model_completions = ModelCompletions(
            text=total_reply, usage=item.usage, resources=total_resources.copy(), succeed=item.succeed
        )
        await hook_manager.run(HookType.AFTER_MODEL_COMPLETION, final_model_completions)

        # 提取挂钩函数的可能的 resources 资源
        new_resources = [r for r in final_model_completions.resources if r not in total_resources]

        # yield 新资源
        for r in new_resources:
            yield ModelStreamCompletions(resources=[r])

        message.respond = total_reply
        message.usage = item.usage

        await hook_manager.run(HookType.ON_FINISHING_CHAT, message)

        if item.succeed:
            await self.database.add_item(message)

    async def refresh(self, userid: str) -> Union[AsyncGenerator[ModelStreamCompletions, None], ModelCompletions]:
        """
        刷新对话

        :userid: 用户唯一标识id
        """
        logger.info(f"用户 {userid} 请求刷新")

        user_history = await self.database.get_user_history(userid, limit=1)

        if not user_history:
            logger.warning("用户对话数据不存在，拒绝刷新")
            return ModelCompletions("你都还没和我说过一句话呢，得和我至少聊上一段才能刷新哦")

        last_item = user_history[0]

        await self.database.mark_history_as_unavailable(userid, 1)

        if not self.model_config.stream:
            return await self.ask(last_item)

        return self.ask_stream(last_item)

    async def reset(self, userid: str) -> str:
        """
        清空历史对话（将用户对话历史记录标记为不可用）
        """
        await self.database.mark_history_as_unavailable(userid)
        return "已成功移除对话历史~"

    async def undo(self, userid: str) -> str:
        await self.database.mark_history_as_unavailable(userid, 1)
        return "已成功撤销上一段对话~"



"""
实现插件的加载和管理

Attributes:
    _plugins (Dict[str, Plugin]): 插件注册表，存储已加载的插件
Functions:
    load_plugin: 加载单个插件
    load_plugins: 加载指定目录下的所有插件
    get_plugins: 获取已加载的插件列表
"""

import importlib
import inspect
import os
from pathlib import Path
from typing import Dict, Optional, Set

import nonebot_plugin_localstore as store
from nonebot import logger

from .models import Plugin
from .utils import path_to_module_name

_plugins: Dict[str, Plugin] = {}
"""插件注册表"""
_declared_plugins: Set[str] = set()
"""已声明插件注册表（不一定加载成功）"""


def load_plugin(plugin_path: Path | str, base_path=Path.cwd()) -> Optional[Plugin]:
    """
    加载单个插件

    :param plugins_dirs: 插件路径
    :param base_path: 外部插件的基准路径
    :return: 插件对象集合
    """
    try:
        logger.info(f"加载插件: {plugin_path}")
        if isinstance(plugin_path, Path):
            plugin_path = path_to_module_name(plugin_path, base_path)

        if plugin_path in _declared_plugins:
            raise ValueError(f"插件 {plugin_path} 包名出现冲突！")
        _declared_plugins.add(plugin_path)

        module = importlib.import_module(plugin_path)
        plugin = Plugin(name=module.__name__.split(".")[-1], module=module, package_name=plugin_path)

        _plugins[plugin.package_name] = plugin

        return plugin

    except Exception as e:
        logger.error(f"加载插件 {plugin_path} 失败: {e}")
        return None


def load_plugins(*plugins_dirs: Path | str, base_path=Path.cwd()) -> set[Plugin]:
    """
    加载传入插件目录中的所有插件

    :param plugins_dirs: 插件目录
    :param base_path: 外部插件的基准路径
    :return: 插件对象集合
    """

    plugins = set()

    for plugin_dir in plugins_dirs:
        plugin_dir_path = Path(plugin_dir) if isinstance(plugin_dir, str) else plugin_dir

        for plugin in os.listdir(plugin_dir_path):
            plugin_path = Path(os.path.join(plugin_dir_path, plugin))
            module_name = None

            if plugin_path.is_file() and plugin_path.suffix == ".py":
                module_name = path_to_module_name(plugin_path.with_suffix(""), base_path)
            elif plugin_path.is_dir() and (plugin_path / Path("__init__.py")).exists():
                module_name = path_to_module_name(plugin_path, base_path)
            if module_name and (loaded_plugin := load_plugin(module_name)):
                plugins.add(loaded_plugin)

    return plugins


def _get_caller_plugin_name() -> Optional[str]:
    """
    获取当前调用插件名
    """
    current_frame = inspect.currentframe()
    if current_frame is None:
        return None

    # find plugin
    frame = current_frame
    while frame := frame.f_back:  # type:ignore
        module_name = (module := inspect.getmodule(frame)) and module.__name__

        if module_name is None:
            return None

        # skip muicebot it self
        package_name = module_name.split(".", maxsplit=1)[0]
        if package_name == "muicebot" and not module_name.startswith("muicebot.builtin_plugins"):
            continue

        # 将模块路径拆解为层级列表（例如 a.b.c → ["a", "a.b", "a.b.c"]）
        module_segments = module_name.split(".")
        candidate_paths = [".".join(module_segments[: i + 1]) for i in range(len(module_segments))]

        # 从长到短查找最长匹配
        for candidate in reversed(candidate_paths):
            if candidate in _declared_plugins:
                return candidate.split(".")[-1]

    return None


def get_plugins() -> Dict[str, Plugin]:
    """
    获取插件列表
    """
    return _plugins


def get_plugin_by_module_name(module_name: str) -> Optional[Plugin]:
    """
    通过包名获取插件对象
    """
    return _plugins.get(module_name, None)


def get_plugin_data_dir() -> Path:
    """
    获取 Muicebot 插件数据目录

    对于 Muicebot 的插件，它们的插件目录位于 Muicebot 的插件目录中下的 `plugins` 文件夹，并以插件名命名
    (`nonebot_plugin_localstore.get_plugin_data_dir`)
    """
    plugin_name = _get_caller_plugin_name()
    plugin_name = plugin_name or ".unknown"

    plugin_dir = store.get_plugin_data_dir() / "plugin"
    plugin_dir = plugin_dir.joinpath(plugin_name).resolve()
    plugin_dir.mkdir(parents=True, exist_ok=True)

    logger.debug(plugin_dir)

    return plugin_dir



"""
模型所需第三方库依赖检查
"""

import importlib

MODEL_DEPENDENCY_MAP = {
    "Azure": ["azure-ai-inference>=1.0.0b7"],
    "Dashscope": ["dashscope>=1.22.1"],
    "Gemini": ["google-genai>=1.8.0"],
    "Ollama": ["ollama>=0.4.7"],
    "Openai": ["openai>=1.64.0"],
}


def get_missing_dependencies(dependencies: list[str]) -> list[str]:
    """
    获取遗失的依赖
    """
    missing = []
    for dep in dependencies:
        try:
            importlib.import_module(dep)
        except ImportError:
            missing.append(dep)
    return missing



from typing import AsyncGenerator, List, Literal, Optional, Union, overload

from google import genai
from google.genai import errors
from google.genai.types import (
    Content,
    ContentOrDict,
    GenerateContentConfig,
    GoogleSearch,
    HarmBlockThreshold,
    HarmCategory,
    Part,
    SafetySetting,
    Tool,
)
from httpx import ConnectError
from nonebot import logger

from ..models import Resource
from ._types import (
    BasicModel,
    ModelCompletions,
    ModelConfig,
    ModelRequest,
    ModelStreamCompletions,
)
from .utils.images import get_file_base64
from .utils.tools import function_call_handler


class Gemini(BasicModel):
    def __init__(self, model_config: ModelConfig) -> None:
        super().__init__(model_config)
        self._require("model_name", "api_key")

        self.model_name = self.config.model_name
        self.api_key = self.config.api_key
        self.enable_search = self.config.online_search

        self.client = genai.Client(api_key=self.api_key)

        self.gemini_config = GenerateContentConfig(
            temperature=self.config.temperature,
            top_p=self.config.top_p,
            top_k=self.config.top_k,
            max_output_tokens=self.config.max_tokens,
            presence_penalty=self.config.presence_penalty,
            frequency_penalty=self.config.frequency_penalty,
            response_modalities=[m.upper() for m in self.config.modalities if m in {"image", "text"}],
            safety_settings=(
                [
                    SafetySetting(
                        category=HarmCategory.HARM_CATEGORY_HARASSMENT,
                        threshold=HarmBlockThreshold.BLOCK_MEDIUM_AND_ABOVE,
                    ),
                    SafetySetting(
                        category=HarmCategory.HARM_CATEGORY_HATE_SPEECH,
                        threshold=HarmBlockThreshold.BLOCK_MEDIUM_AND_ABOVE,
                    ),
                    SafetySetting(
                        category=HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT,
                        threshold=HarmBlockThreshold.BLOCK_MEDIUM_AND_ABOVE,
                    ),
                    SafetySetting(
                        category=HarmCategory.HARM_CATEGORY_CIVIC_INTEGRITY,
                        threshold=HarmBlockThreshold.BLOCK_MEDIUM_AND_ABOVE,
                    ),
                ]
                if self.config.content_security
                else []
            ),
        )

        self.model = self.client.chats.create(model=self.model_name, config=self.gemini_config)

    def __build_user_parts(self, request: ModelRequest) -> list[Part]:
        user_parts: list[Part] = [Part.from_text(text=request.prompt)]

        if not request.resources:
            return user_parts

        for resource in request.resources:
            if resource.type == "image" and resource.path is not None:
                resource.ensure_mimetype()
                user_parts.append(
                    Part.from_bytes(
                        data=get_file_base64(resource.path), mime_type=resource.mimetype or "image/jpeg"  # type:ignore
                    )
                )

        return user_parts

    def __build_tools_list(self, tools: Optional[List] = []):
        format_tools = []

        for tool in tools if tools else []:
            tool = tool["function"]
            required_parameters = tool["required"]
            del tool["required"]
            tool["parameters"]["required"] = required_parameters
            format_tools.append(tool)

        function_tools = Tool(function_declarations=format_tools)  # type:ignore

        if self.enable_search:
            function_tools.google_search = GoogleSearch()

        if tools or self.enable_search:
            self.gemini_config.tools = [function_tools]

    def _build_messages(self, request: ModelRequest) -> list[ContentOrDict]:
        messages: List[ContentOrDict] = []

        if request.history:
            for index, item in enumerate(request.history):
                messages.append(
                    Content(
                        role="user", parts=self.__build_user_parts(ModelRequest(item.message, resources=item.resources))
                    )
                )
                messages.append(Content(role="model", parts=[Part.from_text(text=item.respond)]))

        messages.append(Content(role="user", parts=self.__build_user_parts(request)))

        return messages

    async def _ask_sync(self, messages: list[ContentOrDict], **kwargs) -> ModelCompletions:
        compltions = ModelCompletions()

        try:
            chat = self.client.aio.chats.create(model=self.model_name, config=self.gemini_config, history=messages[:-1])
            message = messages[-1].parts  # type:ignore
            response = await chat.send_message(message=message)  # type:ignore
            if response.usage_metadata:
                total_token_count = response.usage_metadata.total_token_count
                self._total_tokens += total_token_count if total_token_count else -1

            if response.text:
                compltions.text = response.text

            if (
                response.candidates
                and response.candidates[0].content
                and response.candidates[0].content.parts
                and response.candidates[0].content.parts[0].inline_data
                and response.candidates[0].content.parts[0].inline_data.data
            ):
                compltions.resources = [
                    Resource(type="image", raw=response.candidates[0].content.parts[0].inline_data.data)
                ]

            if response.function_calls:
                function_call = response.function_calls[0]
                function_name = function_call.name
                function_args = function_call.args

                logger.info(f"function call 请求 {function_name}, 参数: {function_args}")
                function_return = await function_call_handler(function_name, function_args)  # type:ignore
                logger.success(f"Function call 成功，返回: {function_return}")

                function_response_part = Part.from_function_response(
                    name=function_name,  # type:ignore
                    response={"result": function_return},
                )

                messages.append(Content(role="model", parts=[Part(function_call=function_call)]))
                messages.append(Content(role="user", parts=[function_response_part]))

                return await self._ask_sync(messages)

            compltions.text = compltions.text or "（警告：模型无输出！）"
            compltions.usage = self._total_tokens
            return compltions

        except errors.APIError as e:
            error_message = f"API 状态异常: {e.code}({e.response})"
            compltions.text = error_message
            compltions.succeed = False
            logger.error(error_message)
            logger.error(e.message)
            return compltions

        except ConnectError:
            error_message = "模型加载器连接超时"
            compltions.text = error_message
            compltions.succeed = False
            logger.error(error_message)
            return compltions

    async def _ask_stream(self, messages: list, **kwargs) -> AsyncGenerator[ModelStreamCompletions, None]:
        try:
            total_tokens = 0
            stream = await self.client.aio.models.generate_content_stream(
                model=self.model_name, contents=messages, config=self.gemini_config
            )
            async for chunk in stream:
                stream_completions = ModelStreamCompletions()

                if chunk.text:
                    stream_completions.chunk = chunk.text
                    yield stream_completions

                if chunk.usage_metadata and chunk.usage_metadata.total_token_count:
                    total_tokens = chunk.usage_metadata.total_token_count

                if (
                    chunk.candidates
                    and chunk.candidates[0].content
                    and chunk.candidates[0].content.parts
                    and chunk.candidates[0].content.parts[0].inline_data
                    and chunk.candidates[0].content.parts[0].inline_data.data
                ):
                    stream_completions.resources = [
                        Resource(type="image", raw=chunk.candidates[0].content.parts[0].inline_data.data)
                    ]
                    yield stream_completions

                if chunk.function_calls:
                    function_call = chunk.function_calls[0]
                    function_name = function_call.name
                    function_args = function_call.args

                    logger.info(f"function call 请求 {function_name}, 参数: {function_args}")
                    function_return = await function_call_handler(function_name, function_args)  # type:ignore
                    logger.success(f"Function call 成功，返回: {function_return}")

                    function_response_part = Part.from_function_response(
                        name=function_name,  # type:ignore
                        response={"result": function_return},
                    )

                    messages.append(Content(role="model", parts=[Part(function_call=function_call)]))
                    messages.append(Content(role="user", parts=[function_response_part]))

                    async for final_chunk in self._ask_stream(messages):
                        yield final_chunk

            totaltokens_completions = ModelStreamCompletions()

            self._total_tokens += total_tokens
            totaltokens_completions.usage = self._total_tokens
            yield totaltokens_completions

        except errors.APIError as e:
            stream_completions = ModelStreamCompletions()
            error_message = f"API 状态异常: {e.code}({e.response})"
            stream_completions.chunk = error_message
            logger.error(error_message)
            logger.error(e.message)
            stream_completions.succeed = False
            yield stream_completions
            return

        except ConnectError:
            stream_completions = ModelStreamCompletions()
            error_message = "模型加载器连接超时"
            stream_completions.chunk = error_message
            logger.error(error_message)
            stream_completions.succeed = False
            yield stream_completions
            return

    @overload
    async def ask(self, request: ModelRequest, *, stream: Literal[False] = False) -> ModelCompletions: ...

    @overload
    async def ask(
        self, request: ModelRequest, *, stream: Literal[True] = True
    ) -> AsyncGenerator[ModelStreamCompletions, None]: ...

    async def ask(
        self, request: ModelRequest, *, stream: bool = False
    ) -> Union[ModelCompletions, AsyncGenerator[ModelStreamCompletions, None]]:
        self._total_tokens = 0
        self.__build_tools_list(request.tools)
        self.gemini_config.system_instruction = request.system

        messages = self._build_messages(request)

        if stream:
            return self._ask_stream(messages)

        return await self._ask_sync(messages)



from ._types import HookType
from .manager import (
    hook_manager,
    on_after_completion,
    on_before_completion,
    on_before_pretreatment,
    on_finish_chat,
    on_stream_chunk,
)

__all__ = [
    "HookType",
    "hook_manager",
    "on_after_completion",
    "on_before_completion",
    "on_before_pretreatment",
    "on_finish_chat",
    "on_stream_chunk",
]



import base64
import json
from io import BytesIO
from typing import AsyncGenerator, List, Literal, Union, overload

import openai
from nonebot import logger
from openai import NOT_GIVEN, NotGiven
from openai.types.chat import ChatCompletionMessage, ChatCompletionToolParam

from ..models import Resource
from ._types import (
    BasicModel,
    ModelCompletions,
    ModelConfig,
    ModelRequest,
    ModelStreamCompletions,
)
from .utils.images import get_file_base64
from .utils.tools import function_call_handler


class Openai(BasicModel):
    _tools: List[ChatCompletionToolParam]
    modalities: Union[List[Literal["text", "audio"]], NotGiven]

    def __init__(self, model_config: ModelConfig) -> None:
        super().__init__(model_config)
        self._require("api_key", "model_name")
        self.api_key = self.config.api_key
        self.model = self.config.model_name
        self.api_base = self.config.api_host or "https://api.openai.com/v1"
        self.max_tokens = self.config.max_tokens
        self.temperature = self.config.temperature
        self.stream = self.config.stream
        self.modalities = [m for m in self.config.modalities if m in {"text", "audio"}] or NOT_GIVEN  # type:ignore
        self.audio = self.config.audio if (self.modalities and self.config.audio) else NOT_GIVEN
        self.extra_body = self.config.extra_body

        self.client = openai.AsyncOpenAI(api_key=self.api_key, base_url=self.api_base, timeout=30)
        self._tools = []

    def __build_multi_messages(self, request: ModelRequest) -> dict:
        """
        构建多模态类型

        此模型加载器支持的多模态类型: `audio` `image` `video` `file`
        """
        user_content: List[dict] = [{"type": "text", "text": request.prompt}]

        for resource in request.resources:
            if resource.path is None:
                continue

            elif resource.type == "audio":
                file_format = resource.path.split(".")[-1]
                file_data = f"data:;base64,{get_file_base64(local_path=resource.path)}"
                user_content.append({"type": "input_audio", "input_audio": {"data": file_data, "format": file_format}})

            elif resource.type == "image":
                file_format = resource.path.split(".")[-1]
                file_data = f"data:image/{file_format};base64,{get_file_base64(local_path=resource.path)}"
                user_content.append({"type": "image_url", "image_url": {"url": file_data}})

            elif resource.type == "video":
                file_format = resource.path.split(".")[-1]
                file_data = f"data:;base64,{get_file_base64(local_path=resource.path)}"
                user_content.append({"type": "video_url", "video_url": {"url": file_data}})

            elif resource.type == "file":
                file_format = resource.path.split(".")[-1]
                file_data = f"data:;base64,{get_file_base64(local_path=resource.path)}"
                user_content.append({"type": "file", "file": {"file_data": file_data}})

        return {"role": "user", "content": user_content}

    def _build_messages(self, request: ModelRequest) -> list:
        messages = []

        if request.system:
            messages.append({"role": "system", "content": request.system})

        if request.history:
            for index, item in enumerate(request.history):
                user_content = (
                    {"role": "user", "content": item.message}
                    if not all([item.resources, self.config.multimodal])
                    else self.__build_multi_messages(ModelRequest(item.message, resources=item.resources))
                )

                messages.append(user_content)
                messages.append({"role": "assistant", "content": item.respond})

        user_content = (
            {"role": "user", "content": request.prompt}
            if not request.resources
            else self.__build_multi_messages(request)
        )

        messages.append(user_content)

        return messages

    def _tool_call_request_precheck(self, message: ChatCompletionMessage) -> bool:
        """
        工具调用请求预检
        """
        # We expect a single tool call
        if not (message.tool_calls and len(message.tool_calls) == 1):
            return False

        # We expect the tool to be a function call
        tool_call = message.tool_calls[0]
        if tool_call.type != "function":
            return False

        return True

    async def _ask_sync(self, messages: list, **kwargs) -> ModelCompletions:
        completions = ModelCompletions()

        try:
            response = await self.client.chat.completions.create(
                audio=self.audio,
                model=self.model,
                modalities=self.modalities,
                messages=messages,
                max_tokens=self.max_tokens,
                temperature=self.temperature,
                stream=False,
                tools=self._tools,
                extra_body=self.extra_body,
            )

            result = ""
            message = response.choices[0].message  # type:ignore
            self._total_tokens += response.usage.total_tokens if response.usage else -1

            if (
                hasattr(message, "reasoning_content")  # type:ignore
                and message.reasoning_content  # type:ignore
            ):
                result += f"<think>{message.reasoning_content}</think>"  # type:ignore

            if response.choices[0].finish_reason == "tool_calls" and self._tool_call_request_precheck(
                response.choices[0].message
            ):
                messages.append(response.choices[0].message)
                tool_call = response.choices[0].message.tool_calls[0]  # type:ignore
                arguments = json.loads(tool_call.function.arguments.replace("'", '"'))
                logger.info(f"function call 请求 {tool_call.function.name}, 参数: {arguments}")
                function_return = await function_call_handler(tool_call.function.name, arguments)
                logger.success(f"Function call 成功，返回: {function_return}")
                messages.append(
                    {
                        "tool_call_id": tool_call.id,
                        "role": "tool",
                        "name": tool_call.function.name,
                        "content": function_return,
                    }
                )
                return await self._ask_sync(messages)

            if message.content:  # type:ignore
                result += message.content  # type:ignore

            # 多模态消息处理（目前仅支持 audio 输出）
            if response.choices[0].message.audio:
                wav_bytes = base64.b64decode(response.choices[0].message.audio.data)
                completions.resources = [Resource(type="audio", raw=wav_bytes)]

            completions.text = result or "（警告：模型无输出！）"
            completions.usage = self._total_tokens

        except openai.APIConnectionError as e:
            error_message = f"API 连接错误: {e}"
            completions.text = error_message
            logger.error(error_message)
            logger.error(e.__cause__)
            completions.succeed = False

        except openai.APIStatusError as e:
            error_message = f"API 状态异常: {e.status_code}({e.response})"
            completions.text = error_message
            logger.error(error_message)
            completions.succeed = False

        return completions

    async def _ask_stream(self, messages: list, **kwargs) -> AsyncGenerator[ModelStreamCompletions, None]:
        is_insert_think_label = False
        function_id = ""
        function_name = ""
        function_arguments = ""
        audio_string = ""

        try:
            response = await self.client.chat.completions.create(
                audio=self.audio,
                model=self.model,
                modalities=self.modalities,
                messages=messages,
                max_tokens=self.max_tokens,
                temperature=self.temperature,
                stream=True,
                stream_options={"include_usage": True},
                tools=self._tools,
                extra_body=self.extra_body,
            )

            async for chunk in response:
                stream_completions = ModelStreamCompletions()

                # 获取 usage （最后一个包中返回）
                if chunk.usage:
                    self._total_tokens += chunk.usage.total_tokens
                    stream_completions.usage = self._total_tokens

                if not chunk.choices:
                    yield stream_completions
                    continue

                # 处理 Function call
                if chunk.choices[0].delta.tool_calls:
                    tool_call = chunk.choices[0].delta.tool_calls[0]
                    if tool_call.id:
                        function_id = tool_call.id
                    if tool_call.function:
                        if tool_call.function.name:
                            function_name += tool_call.function.name
                        if tool_call.function.arguments:
                            function_arguments += tool_call.function.arguments

                delta = chunk.choices[0].delta
                answer_content = delta.content

                # 处理思维过程 reasoning_content
                if (
                    hasattr(delta, "reasoning_content") and delta.reasoning_content  # type:ignore
                ):
                    reasoning_content = chunk.choices[0].delta.reasoning_content  # type:ignore
                    stream_completions.chunk = (
                        reasoning_content if is_insert_think_label else "<think>" + reasoning_content
                    )
                    yield stream_completions
                    is_insert_think_label = True

                elif answer_content:
                    stream_completions.chunk = (
                        answer_content if not is_insert_think_label else "</think>" + answer_content
                    )
                    yield stream_completions
                    is_insert_think_label = False

                # 处理多模态消息 (audio-only) (非标准方法，可能出现问题)
                if hasattr(chunk.choices[0].delta, "audio"):
                    audio = chunk.choices[0].delta.audio  # type:ignore
                    if audio.get("data", None):
                        audio_string += audio.get("data")
                    stream_completions.chunk = audio.get("transcript", "")
                    yield stream_completions

            if function_id:
                logger.info(f"function call 请求 {function_name}, 参数: {function_arguments}")
                function_return = await function_call_handler(function_name, json.loads(function_arguments))
                logger.success(f"Function call 成功，返回: {function_return}")
                messages.append(
                    {
                        "role": "assistant",
                        "content": None,
                        "tool_calls": [
                            {
                                "id": function_id,
                                "type": "function",
                                "function": {"name": function_name, "arguments": function_arguments},
                            }
                        ],
                    }
                )
                messages.append(
                    {
                        "tool_call_id": function_id,
                        "role": "tool",
                        "content": function_return,
                    }
                )

                async for chunk in self._ask_stream(messages):
                    yield chunk

            # 处理多模态返回
            if audio_string:
                import numpy as np
                import soundfile as sf

                wav_bytes = base64.b64decode(audio_string)
                pcm_data = np.frombuffer(wav_bytes, dtype=np.int16)
                wav_io = BytesIO()
                sf.write(wav_io, pcm_data, samplerate=24000, format="WAV")

                stream_completions = ModelStreamCompletions()
                stream_completions.resources = [Resource(type="audio", raw=wav_io)]
                yield stream_completions

        except openai.APIConnectionError as e:
            error_message = f"API 连接错误: {e}"
            logger.error(error_message)
            logger.error(e.__cause__)
            stream_completions = ModelStreamCompletions()
            stream_completions.chunk = error_message
            stream_completions.succeed = False
            yield stream_completions

        except openai.APIStatusError as e:
            error_message = f"API 状态异常: {e.status_code}({e.response})"
            logger.error(error_message)
            stream_completions = ModelStreamCompletions()
            stream_completions.chunk = error_message
            stream_completions.succeed = False
            yield stream_completions

    @overload
    async def ask(self, request: ModelRequest, *, stream: Literal[False] = False) -> ModelCompletions: ...

    @overload
    async def ask(
        self, request: ModelRequest, *, stream: Literal[True] = True
    ) -> AsyncGenerator[ModelStreamCompletions, None]: ...

    async def ask(
        self, request: ModelRequest, *, stream: bool = False
    ) -> Union[ModelCompletions, AsyncGenerator[ModelStreamCompletions, None]]:
        self._tools = request.tools if request.tools else NOT_GIVEN  # type:ignore
        self._total_tokens = 0

        messages = self._build_messages(request)

        if stream:
            return self._ask_stream(messages)

        return await self._ask_sync(messages)



import json
from pathlib import Path
from typing import Any, Dict

from pydantic import BaseModel, Field

CONFIG_PATH = Path("./configs/mcp.json")


class mcpServer(BaseModel):
    command: str
    """执行指令"""
    args: list = Field(default_factory=list)
    """命令参数"""
    env: dict[str, Any] = Field(default_factory=dict)
    """环境配置"""


mcpConfig = Dict[str, mcpServer]


def get_mcp_server_config() -> mcpConfig:
    """
    从 MCP 配置文件 `config/mcp.json` 中获取 MCP Server 配置
    """
    if not CONFIG_PATH.exists():
        return {}

    with open(CONFIG_PATH, "r", encoding="utf-8") as f:
        configs = json.load(f) or {}

    mcp_config: mcpConfig = dict()

    for name, srv_config in configs["mcpServers"].items():
        mcp_config[name] = mcpServer(**srv_config)

    return mcp_config


server_config = get_mcp_server_config()



"""
定义 Function_call 插件的参数类
"""

from typing import Any

from pydantic import BaseModel, Field


class Parameter(BaseModel):
    """
    Function_call 插件参数对象
    """

    type: str
    """参数类型描述 string integer等"""
    description: str
    """参数描述"""
    default: Any = None
    """默认值"""
    properties: dict[str, Any] = {}
    """参数定义属性，例如最大值最小值等"""
    required: bool = False
    """是否必须"""

    def data(self) -> dict[str, Any]:
        """
        生成参数描述信息

        :return: 可用于 Function_call 的字典
        """
        return {
            "type": self.type,
            "description": self.description,
            **{key: value for key, value in self.properties.items() if value is not None},
        }


class ParamTypes:
    STRING = "string"
    INTEGER = "integer"
    ARRAY = "array"
    OBJECT = "object"
    BOOLEAN = "boolean"
    NUMBER = "number"


class String(Parameter):
    type: str = ParamTypes.STRING
    properties: dict[str, Any] = Field(default_factory=dict)
    enum: list[str] | None = None


class Integer(Parameter):
    type: str = ParamTypes.INTEGER
    properties: dict[str, Any] = Field(default_factory=lambda: {"minimum": 0, "maximum": 100})

    minimum: int | None = None
    maximum: int | None = None


class Array(Parameter):
    type: str = ParamTypes.ARRAY
    properties: dict[str, Any] = Field(default_factory=lambda: {"items": {"type": "string"}})
    items: str = Field("string", description="数组元素类型")



from types import ModuleType
from typing import Any

from pydantic import BaseModel


class PluginMetadata(BaseModel):
    """MuiceBot 插件元数据"""

    name: str
    """插件名"""
    description: str
    """插件描述"""
    usage: str
    """插件用法"""
    homepage: str | None = None
    """(可选) 插件主页，通常为开源存储库地址"""
    config: type[BaseModel] | None = None
    """插件配置项类，如无需配置可不填写"""
    extra: dict[Any, Any] | None = None
    """不知道干嘛的 extra 信息，我至今都没搞懂，喜欢的可以填"""


class Plugin(BaseModel):
    """MuiceBot 插件对象"""

    name: str
    """插件名称"""
    module: ModuleType
    """插件模块对象"""
    package_name: str
    """模块包名"""

    def __hash__(self) -> int:
        return hash(self.package_name)

    def __eq__(self, other: Any) -> bool:
        return self.package_name == other.package_name if hasattr(other, "package_name") else False

    def __str__(self) -> str:
        return self.package_name

    class Config:
        arbitrary_types_allowed = True



import importlib
from functools import lru_cache

ADAPTER_CLASSES = {}
"""动态适配器注册表"""


@lru_cache()
def safe_import(path: str):
    """
    安全导入：即使导入出现问题也不会出现报错
    """
    try:
        module_path, class_name = path.rsplit(".", 1)
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except ImportError:
        return None


ADAPTER_CLASSES = {
    "onebot_v12": safe_import("nonebot.adapters.onebot.v12.Bot"),
    "UnsupportedParam": safe_import("nonebot.adapters.onebot.v12.exception.UnsupportedParam"),
    "onebot_v11": safe_import("nonebot.adapters.onebot.v11.Bot"),
    "telegram_event": safe_import("nonebot.adapters.telegram.Event"),
    "telegram_file": safe_import("nonebot.adapters.telegram.message.File"),
    "qq_event": safe_import("nonebot.adapters.telegram.qq.Event"),
}
